"""
Subpackage ``fits`` includes fits handling functions:
    - fits opening
    - fits info
    - fits writing
    - appending extensions to a fit file
    - ADI cube opening (cube with PA attached as HDU extension)
"""
from __future__ import absolute_import

from .fits import *
from .vipds9 import *




"""
The subpackage ``calib`` contains some useful cosmetics and pre-processing
functionalities:
    - resizing frames and cubes : upscaling/pixel binning,
    - shifting frames,
    - rotating frames and cubes,
    - cropping frames and cubes,
    - removing bad pixels from frames and cubes,
    - correcting nan values from frames and cubes,
    - detecting bad frames in cubes, using:
        - pixel statistics in annulus or circular aperture,
        - ellipticity of a point like source,
        - frames correlation,
    - temporal subsampling of cubes (mean, median, trimmed mean),
    - registration (re-centering) of frames, using:
        - centroid fitting a 2d gaussian or moffat,
        - DFT upsampling or fourier cross-correlation (Guizar et al. 2008),
        - radon transform for broadband frames (Pueyo et al. 2014),
        - using satellite/waffle spots (fitting plus intersection).
    - sky subtraction (PCA method).

Astronomical calibration functionality like flat fielding and dark-sky
subtraction, in spite of its simplicity was not included in VIP because of the
heterogeneity of the datasets coming from different observatories (each having
different data storage and headers). You can perform this in python in procedures
of a few lines or using dedicated instrument pipelines such as esorex (ESO
instruments)."""
from __future__ import absolute_import

from .badframes import *
from .badpixremoval import *
from .cosmetics import *
from .derotation import *
from .parangles import *
from .recentering import *
from .rescaling import *
from .skysubtraction import *
from .subsampling import *





#! /usr/bin/env python

"""
LLSG (Gomez Gonzalez et al. 2016)
"""
from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['llsg']


import numpy as np
from scipy.linalg import qr
import itertools as itt
from multiprocessing import Pool, cpu_count
from astropy.stats import median_absolute_deviation
from ..conf import time_ini, timing
from ..preproc import cube_derotate, cube_collapse
from ..var import get_annulus_segments, cube_filter_highpass
from ..pca.svd import svd_wrapper, get_eigenvectors
from .thresholding import thresholding
from ..conf.utils_conf import eval_func_tuple as EFT


cube_init = None


def llsg(cube, angle_list, fwhm, rank=10, thresh=1, max_iter=10,
         low_rank_ref=False, low_rank_mode='svd', auto_rank_mode='noise',
         residuals_tol=1e-1, cevr=0.9, thresh_mode='soft', nproc=1,
         asize=None, n_segments=4, azimuth_overlap=None, radius_int=None,
         random_seed=None, imlib='opencv', interpolation='lanczos4',
         high_pass=None, collapse='median', full_output=False, verbose=True,
         debug=False):
    """ Local Low-rank plus Sparse plus Gaussian-noise decomposition (LLSG) as
    described in Gomez Gonzalez et al. 2016. This first version of our algorithm
    aims at decomposing ADI cubes into three terms L+S+G (low-rank, sparse and
    Gaussian noise). Separating the noise from the S component (where the moving
    planet should stay) allow us to increase the SNR of potential planets.

    The three tunable parameters are the *rank* or expected rank of the L
    component, the *thresh* or threshold for encouraging sparsity in the S
    component and *max_iter* which sets the number of iterations. The rest of
    parameters can be tuned at the users own risk (do it if you know what you're
    doing).

    Parameters
    ----------
    cube : array_like, 3d
        Input ADI cube.
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame.
    fwhm : float
        Known size of the FHWM in pixels to be used.
    rank : int, optional
        Expected rank of the L component.
    thresh : float, optional
        Factor that scales the thresholding step in the algorithm.
    max_iter : int, optional
        Sets the number of iterations.
    low_rank_ref :
        If True the first estimation of the L component is obtained from the
        remaining segments in the same annulus.
    low_rank_mode : {'svd', 'brp'}, optional
        Sets the method of solving the L update.
    auto_rank_mode : {'noise', 'cevr'}, str optional
        If ``rank`` is None, then ``auto_rank_mode`` sets the way that the
        ``rank`` is determined: the noise minimization or the cumulative
        explained variance ratio (when 'svd' is used).
    residuals_tol : float, optional
        The value of the noise decay to be used when ``rank`` is None and
        ``auto_rank_mode`` is set to ``noise``.
    cevr : float, optional
        Float value in the range [0,1] for selecting the cumulative explained
        variance ratio to choose the rank automatically (if ``rank`` is None).
    thresh_mode : {'soft', 'hard'}, optional
        Sets the type of thresholding.
    nproc : None or int, optional
        Number of processes for parallel computing. If None the number of
        processes will be set to (cpu_count()/2). By default the algorithm works
        in single-process mode.
    asize : int or None, optional
        If ``asize`` is None then each annulus will have a width of ``2*asize``.
        If an integer then it is the width in pixels of each annulus.
    n_segments : int or list of ints, optional
        The number of segments for each annulus. When a single integer is given
        it is used for all annuli.
    azimuth_overlap : int or None, optional
        Sets the amount of azimuthal averaging.
    radius_int : int, optional
        The radius of the innermost annulus. By default is 0, if >0 then the
        central circular area is discarded.
    random_seed : int or None, optional
        Controls the seed for the Pseudo Random Number generator.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    high_pass : odd int or None, optional
        If set to an odd integer <=7, a high-pass filter is applied to the
        frames. The ``vip_hci.var.frame_filter_highpass`` is applied twice,
        first with the mode ``median-subt`` and a large window, and then with
        ``laplacian-conv`` and a kernel size equal to ``high_pass``. 5 is an
        optimal value when ``fwhm`` is ~4.
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    full_output: bool, optional
        Whether to return the final median combined image only or with other
        intermediate arrays.
    verbose : bool, optional
        If True prints to stdout intermediate info.
    debug : bool, optional
        Whether to output some intermediate information.

    Returns
    -------
    frame_s : array_like, 2d
        Final frame (from the S component) after rotation and median-combination.

    If ``full_output`` is True, the following intermediate arrays are returned:
    list_l_array_der, list_s_array_der, list_g_array_der, frame_l, frame_s, frame_g

    """
    if not cube.ndim == 3:
        raise TypeError("Input array is not a cube (3d array)")
    if not cube.shape[0] == angle_list.shape[0]:
        msg = "Angle list vector has wrong length. It must equal the number"
        msg += " frames in the cube"
        raise TypeError(msg)

    if low_rank_mode == 'brp':
        if rank is None:
            msg = "Auto rank only works with SVD low_rank_mode."
            msg += " Set a value for the rank parameter"
            raise ValueError(msg)
        if low_rank_ref:
            msg = "Low_rank_ref only works with SVD low_rank_mode"
            raise ValueError(msg)

    global cube_init
    if high_pass is not None:
        cube_init = cube_filter_highpass(cube, 'median-subt', median_size=19,
                                         verbose=False)
        cube_init = cube_filter_highpass(cube_init, 'laplacian-conv',
                                         kernel_size=high_pass, verbose=False)
    else:
        cube_init = cube

    if verbose:
        start_time = time_ini()
    n, y, x = cube.shape

    if azimuth_overlap == 0:
        azimuth_overlap = None

    if radius_int is None:
        radius_int = 0

    if nproc is None:   # Hyper-threading "duplicates" the cores -> cpu_count/2
        nproc = (cpu_count()/2)

    # Same number of pixels per annulus
    if asize is None:
        annulus_width = int(np.ceil(2 * fwhm))  # as in the paper
    elif isinstance(asize, int):
        annulus_width = asize
    n_annuli = int(np.floor((y / 2 - radius_int) / annulus_width))

    if n_segments is None:
        n_segments = [4 for _ in range(n_annuli)]   # as in the paper
    elif isinstance(n_segments, int):
        n_segments = [n_segments for _ in range(n_annuli)]
    elif n_segments == 'auto':
        n_segments = []
        n_segments.append(2)    # for first annulus
        n_segments.append(3)    # for second annulus
        ld = 2 * np.tan(360/4/2) * annulus_width
        for i in range(2, n_annuli):    # rest of annuli
            radius = i * annulus_width
            ang = np.rad2deg(2 * np.arctan(ld / (2 * radius)))
            n_segments.append(int(np.ceil(360/ang)))

    if verbose:
        print('Annuli = {:}'.format(n_annuli))

    # Azimuthal averaging of residuals
    if azimuth_overlap is None:
        azimuth_overlap = 360   # no overlapping, single config of segments
    n_rots = int(360. / azimuth_overlap)

    matrix_s = np.zeros((n_rots, n, y, x))
    if full_output:
        matrix_l = np.zeros((n_rots, n, y, x))
        matrix_g = np.zeros((n_rots, n, y, x))

    # Looping the he annuli
    if verbose:
        print('Processing annulus: ')
    for ann in range(n_annuli):
        inner_radius = radius_int + ann * annulus_width
        n_segments_ann = n_segments[ann]
        if verbose:
            print('{} : in_rad={}, n_segm={}'.format(ann+1, inner_radius,
                                                     n_segments_ann))
        for i in range(n_rots):
            theta_init = i * azimuth_overlap
            indices = get_annulus_segments(cube[0], inner_radius,
                                           annulus_width, n_segments_ann,
                                           theta_init)

            if nproc == 1:
                for j in range(n_segments_ann):
                    yy = indices[j][0]
                    xx = indices[j][1]

                    patch = _decompose_patch(indices, j, n_segments_ann,
                                             rank, low_rank_ref, low_rank_mode,
                                             thresh, thresh_mode, max_iter,
                                             auto_rank_mode, cevr, residuals_tol,
                                             random_seed, debug, full_output)

                    if full_output:
                        matrix_l[i, :, yy, xx] = patch[0]
                        matrix_s[i, :, yy, xx] = patch[1]
                        matrix_g[i, :, yy, xx] = patch[2]
                    else:
                        matrix_s[i, :, yy, xx] = patch

            elif nproc > 1:
                pool = Pool(processes=int(nproc))
                res = pool.map(EFT, zip(itt.repeat(_decompose_patch),
                                        itt.repeat(indices),
                                        range(n_segments_ann),
                                        itt.repeat(n_segments_ann),
                                        itt.repeat(rank),
                                        itt.repeat(low_rank_ref),
                                        itt.repeat(low_rank_mode),
                                        itt.repeat(thresh),
                                        itt.repeat(thresh_mode),
                                        itt.repeat(max_iter),
                                        itt.repeat(auto_rank_mode),
                                        itt.repeat(cevr),
                                        itt.repeat(residuals_tol),
                                        itt.repeat(random_seed),
                                        itt.repeat(debug),
                                        itt.repeat(full_output)))
                patches = np.array(res)
                pool.close()

                for j in range(n_segments_ann):
                    yy = indices[j][0]
                    xx = indices[j][1]

                    if full_output:
                        matrix_l[i, :, yy, xx] = patches[j][0]
                        matrix_s[i, :, yy, xx] = patches[j][1]
                        matrix_g[i, :, yy, xx] = patches[j][2]
                    else:
                        matrix_s[i, :, yy, xx] = patches[j]

            else:
                raise ValueError("nproc must be a positive integer")

    if full_output:
        list_s_array_der = [cube_derotate(matrix_s[k], angle_list, imlib=imlib,
                                          interpolation=interpolation)
                            for k in range(n_rots)]
        list_frame_s = [cube_collapse(list_s_array_der[k], mode=collapse)
                        for k in range(n_rots)]
        frame_s = cube_collapse(np.array(list_frame_s), mode=collapse)

        list_l_array_der = [cube_derotate(matrix_l[k], angle_list, imlib=imlib,
                                          interpolation=interpolation)
                            for k in range(n_rots)]
        list_frame_l = [cube_collapse(list_l_array_der[k], mode=collapse)
                        for k in range(n_rots)]
        frame_l = cube_collapse(np.array(list_frame_l), mode=collapse)

        list_g_array_der = [cube_derotate(matrix_g[k], angle_list, imlib=imlib,
                                          interpolation=interpolation)
                            for k in range(n_rots)]
        list_frame_g = [cube_collapse(list_g_array_der[k], mode=collapse)
                        for k in range(n_rots)]
        frame_g = cube_collapse(np.array(list_frame_g), mode=collapse)

    else:
        list_s_array_der = [cube_derotate(matrix_s[k], angle_list, imlib=imlib,
                                          interpolation=interpolation)
                            for k in range(n_rots)]
        list_frame_s = [cube_collapse(list_s_array_der[k], mode=collapse)
                        for k in range(n_rots)]

        # list_frame_s = []
        # for nr in range(n_rots):
        #     list_frame_s.append(np.mean(list_s_array_der[nr], axis=0) /
        #                         np.std(list_s_array_der[nr], axis=0))

        frame_s = cube_collapse(np.array(list_frame_s), mode=collapse)

    if verbose:
        print('')
        timing(start_time)

    if full_output:
        return(list_l_array_der, list_s_array_der, list_g_array_der,
               frame_l, frame_s, frame_g)
    else:
        return frame_s


def _decompose_patch(indices, i_patch, n_segments_ann, rank, low_rank_ref,
                     low_rank_mode, thresh, thresh_mode, max_iter,
                     auto_rank_mode, cevr, residuals_tol, random_seed,
                     debug=False, full_output=False):
    """ Patch decomposition.
    """
    j = i_patch
    yy = indices[j][0]
    xx = indices[j][1]
    data_segm = cube_init[:, yy, xx]

    if low_rank_ref:
        ref_segments = list(range(n_segments_ann))
        ref_segments.pop(j)
        for m, n in enumerate(ref_segments):
            if m == 0:
                yy_ref = indices[n][0]
                xx_ref = indices[n][1]
            else:
                yy_ref = np.hstack((yy_ref, indices[n][0]))
                xx_ref = np.hstack((xx_ref, indices[n][1]))
        data_ref = cube_init[:, yy_ref, xx_ref]
    else:
        data_ref = data_segm

    patch = _patch_rlrps(data_segm, data_ref, rank, low_rank_ref,
                         low_rank_mode, thresh, thresh_mode,
                         max_iter, auto_rank_mode, cevr,
                         residuals_tol, random_seed, debug=debug,
                         full_output=full_output)
    return patch


def _patch_rlrps(array, array_ref, rank, low_rank_ref, low_rank_mode,
                 thresh, thresh_mode, max_iter, auto_rank_mode='noise',
                 cevr=0.9, residuals_tol=1e-2, random_seed=None, debug=False,
                 full_output=False):
    """ Patch decomposition based on GoDec/SSGoDec (Zhou & Tao 2011)
    """
    ############################################################################
    # Initializing L and S
    ############################################################################
    L = array
    if low_rank_ref: L_ref = array_ref.T
    else: L_ref = None
    S = np.zeros_like(L)
    random_state = np.random.RandomState(random_seed)
    itr = 0
    power = 0
    svdlib = 'randsvd'

    while itr <= max_iter:
        ########################################################################
        # Updating L
        ########################################################################
        if low_rank_mode == 'brp':
            Y2 = random_state.randn(L.shape[1], rank)
            for _ in range(power + 1):
                Y1 = np.dot(L, Y2)
                Y2 = np.dot(L.T, Y1)
            Q, _ = qr(Y2, mode='economic')
            Lnew = np.dot(np.dot(L, Q), Q.T)

        elif low_rank_mode == 'svd':
            if itr == 0:
                PC = get_eigenvectors(rank, L, svdlib, mode=auto_rank_mode,
                                      cevr=cevr, noise_error=residuals_tol,
                                      data_ref=L_ref, debug=debug)
                rank = PC.shape[0]  # so we can use the optimized rank
                if low_rank_ref:
                    Lnew = np.dot(np.dot(PC, L).T, PC).T
                else:
                    Lnew = np.dot(np.dot(L, PC.T), PC)
            else:
                rank_i = min(rank, min(L.shape[0], L.shape[1]))
                PC = svd_wrapper(L, svdlib, rank_i, False, False,
                                 random_state=random_state)
                Lnew = np.dot(np.dot(L, PC.T), PC)

        else:
            raise RuntimeError('Low Rank estimation mode not recognized.')

        ########################################################################
        # Updating S
        ########################################################################
        T = L - Lnew + S
        threshold = np.sqrt(median_absolute_deviation(T.ravel())) * thresh

        # threshold = np.sqrt(median_absolute_deviation(T, axis=0)) * thresh
        # threshmat = np.zeros_like(T)
        # for i in range(threshmat.shape[0]):
        #     threshmat[i] = threshold
        # threshold = threshmat

        if debug:
            print('threshold = {:.3f}'.format(threshold))
        S = thresholding(T, threshold, thresh_mode)

        T -= S
        L = Lnew + T
        itr += 1

    G = array - L - S

    L = L.T
    S = S.T
    G = G.T

    if full_output:
        return L, S, G
    else:
        return S





#! /usr/bin/env python

"""
Module with pixel and frame subsampling functions.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['cube_collapse',
           'cube_subsample',
           'cube_subsample_trimmean']

import numpy as np


def cube_collapse(cube, mode='median', n=50):
    """ Collapses a cube into a frame (3D array -> 2D array) depending on the
    parameter ``mode``. It's possible to perform a trimmed mean combination of the
    frames based on description in Brandt+ 2012.
    
    Parameters
    ----------
    cube : array_like
        Cube.
    mode : {'median', 'mean', 'sum', 'trimmean', 'max'}, str optional
        Sets the way of collapsing the images in the cube.
    n : int, optional
        Sets the discarded values at high and low ends. When n = N is the same
        as taking the mean, when n = 1 is like taking the median.
        
    Returns
    -------
    frame : array_like
        Output array, cube combined. 
    """
    arr = cube
    if not arr.ndim == 3:
        raise TypeError('The input array is not a cube or 3d array.')
    
    if mode == 'mean':
        frame = np.mean(arr, axis=0)
    elif mode == 'median':
        frame = np.median(arr, axis=0)
    elif mode == 'sum':
        frame = np.sum(arr, axis=0)
    elif mode == 'max':
        frame = np.max(arr, axis=0)
    elif mode == 'trimmean':
        N = arr.shape[0]
        if N % 2 == 0:
            k = (N - n)//2
        else:
            k = (N - n)/2                                                               
        
        frame = np.empty_like(arr[0])                                    
        for index, _ in np.ndenumerate(arr[0]):
            sort = np.sort(arr[:, index[0], index[1]])
            frame[index] = np.mean(sort[k:N-k])
            
    return frame


def cube_subsample(array, n, mode="mean", parallactic=None, verbose=True):
    """Mean/Median combines frames in cube with window n.
    
    Parameters
    ----------
    n : int
        Window for mean/median.
    array : array_like 
        Input 3d array, cube.
    mode : {'mean','median'}
        Switch for choosing mean or median.
    parallactic: array_like
        List of corresponding parallactic angles.
        
    Returns
    -------
    arr_view : array_like
        Resulting array.
    angles : array_like
        Parallactic angles.
    """
    if not array.ndim == 3:
        raise TypeError('The input array is not a cube or 3d array.')
    m = int(array.shape[0]/n) 
    resid = array.shape[0]%n                                                       
    y = array.shape[1]
    x = array.shape[2]
    arr = np.empty([m, y, x]) 
    if parallactic is not None:
        angles = np.zeros(m)
        
    if mode == 'median':  func = np.median
    elif mode=='mean':  func = np.mean
    else:  
        raise ValueError('Mode should be either Mean or Median.') 
        
    for i in range(m):
        arr[i, :, :] = func(array[:n, :, :], axis=0) 
        if parallactic is not None:  angles[i] = func(parallactic[:n])
        if i >= 1:
            arr[i, :, :] = func(array[n*i:n*i+n, :, :], axis=0)
            if parallactic is not None:
                angles[i] = func(parallactic[n*i:n*i+n])

    if verbose:
        print("Datacube subsampled by taking the {:} of {:} frames".format(mode ,n))
        if resid > 0:
            msg = "Initial # of frames and window are not multiples ({:} frames were dropped)"
            print(msg.format(resid))     
        print("New cube contains {:} frames".format(m))
                                   
    if parallactic is not None:
        return arr, angles
    else:
        return arr


def cube_subsample_trimmean(arr, n, m):
    """Performs a trimmed mean combination every m frames in a cube. Based on 
    description in Brandt+ 2012.
    
    Parameters
    ----------
    arr : array_like
        Cube.
    n : int
        Sets the discarded values at high and low ends. When n = N is the same
        as taking the mean, when n = 1 is like taking the median.
    m : int
        Window from the trimmed mean.
        
    Returns
    -------
    arr_view : array_like
        Output array, cube combined. 
    """    
    if not arr.ndim == 3:
        raise TypeError('The input array is not a cube or 3d array.')
    num = int(arr.shape[0]/m)
    res = int(arr.shape[0]%m)                                                       
    y = arr.shape[1]
    x = arr.shape[2]
    arr2 = np.empty([num+2, y, x]) 
    for i in range(num):
        arr2[0] = cube_collapse(arr[:m, :, :], 'trimmean', n)                         
        if i > 0:
            arr2[i] = cube_collapse(arr[m*i:m*i+m, :, :], 'trimmean', n)
    arr2[num] = cube_collapse(arr[-res:, :, :], 'trimmean', n)     
    arr_view = arr2[:num+1]                                                      # slicing until m+1 - last index not included
    print("\nDone trimmed mean over FITS-Cube with window m=" + str(m))
    return arr_view





#! /usr/bin/env python

"""
Module with ADI algorithm (median psf subtraction).
"""

from __future__ import division 
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['adi']

import numpy as np
import itertools as itt
from multiprocessing import Pool, cpu_count
from ..conf import time_ini, timing
from ..var import get_annulus, mask_circle
from ..preproc import cube_derotate, cube_collapse, check_pa_vector
from ..conf.utils_conf import eval_func_tuple as EFT
from .adi_utils import _find_indices, _define_annuli


array = None


def adi(cube, angle_list, fwhm=4, radius_int=0, asize=2, delta_rot=1, 
        mode='fullfr', nframes=4, imlib='opencv', interpolation='lanczos4',
        collapse='median', nproc=1, full_output=False, verbose=True):
    """ Algorithm based on Marois et al. 2006 on Angular Differential Imaging.   
    First the median frame is subtracted, then (in annular `mode``) the median
    of the four closest frames taking into account the pa_threshold (field
    rotation).
    
    Parameters
    ----------
    cube : array_like, 3d
        Input cube.
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame.
    fwhm : float
        Known size of the FHWM in pixels to be used. Default is 4.
    radius_int : int, optional
        The radius of the innermost annulus. By default is 0, if >0 then the 
        central circular area is discarded.
    asize : int, optional
        The size of the annuli, in FWHM. Default is 2.
    delta_rot : int, optional
        Factor for increasing the parallactic angle threshold, expressed in FWHM.
        Default is 1 (excludes 1 FHWM on each side of the considered frame).
    mode : {"fullfr","annular"}, str optional
        In "simple" mode only the median frame is subtracted, in "annular" mode
        also the 4 closest frames given a PA threshold (annulus-wise) are 
        subtracted.
    nframes : even int optional
        Number of frames to be used for building the optimized reference PSF 
        when working in annular mode.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    nproc : None or int, optional
        Number of processes for parallel computing. If None the number of
        processes will be set to (cpu_count()/2). By default the algorithm works
        in single-process mode.
    full_output: bool, optional
        Whether to return the final median combined image only or with other 
        intermediate arrays. 
    verbose : bool, optional
        If True prints to stdout intermediate info.
        
    Returns
    -------
    frame : array_like, 2d
        Median combination of the de-rotated cube.
    If full_output is True:  
    cube_out : array_like, 3d
        The cube of residuals.
    cube_der : array_like, 3d
        The derotated cube of residuals.
         
    """
    global array
    array = cube
    
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array')
    if not array.shape[0] == angle_list.shape[0]:
        raise TypeError('Input vector or parallactic angles has wrong length')
    if not nframes % 2 == 0:
        raise TypeError('nframes argument must be even value')
    
    n, y, _ = array.shape
     
    if verbose:
        start_time = time_ini()

    if nproc is None:   # Hyper-threading "duplicates" the cores -> cpu_count/2
        nproc = (cpu_count()/2)

    angle_list = check_pa_vector(angle_list)
    
    # The median frame (basic psf reference) is first subtracted from each frame
    model_psf = np.median(array, axis=0)
    array = array - model_psf

    # Depending on the ``mode``
    if mode == 'fullfr':
        if radius_int > 0:
            cube_out = mask_circle(array, radius_int)
        else:
            cube_out = array
        if verbose:
            print('Median psf reference subtracted')
    
    elif mode == 'annular':
        annulus_width = int(asize * fwhm)  # equal size for all annuli
        n_annuli = int(np.floor((y / 2 - radius_int) / annulus_width))
        if verbose:
            print('N annuli =', n_annuli, ', FWHM =', fwhm, '\n')

        cube_out = np.zeros_like(array)

        if nproc == 1:
            for ann in range(n_annuli):
                mres, yy, xx = _median_subt_ann(ann, angle_list, n_annuli, fwhm,
                                                radius_int, annulus_width,
                                                delta_rot, nframes, verbose)
                cube_out[:, yy, xx] = mres
        elif nproc > 1:
            pool = Pool(processes=int(nproc))
            res = pool.map(EFT, zip(itt.repeat(_median_subt_ann),
                                    range(n_annuli), itt.repeat(angle_list),
                                    itt.repeat(n_annuli), itt.repeat(fwhm),
                                    itt.repeat(radius_int),
                                    itt.repeat(annulus_width),
                                    itt.repeat(delta_rot), itt.repeat(nframes),
                                    itt.repeat(verbose)))
            res = np.array(res)
            pool.close()
            mres = res[:, 0]
            yy = res[:, 1]
            xx = res[:, 2]
            for ann in range(n_annuli):
                cube_out[:, yy[ann], xx[ann]] = mres[ann]

        if verbose:
            print('Optimized median psf reference subtracted')
        
    else:
        raise RuntimeError('Mode not recognized')
    
    cube_der = cube_derotate(cube_out, angle_list, imlib=imlib,
                             interpolation=interpolation)
    frame = cube_collapse(cube_der, mode=collapse)
    if verbose:
        print('Done derotating and combining')
        timing(start_time)
    if full_output:
        return cube_out, cube_der, frame 
    else:
        return frame 


def _median_subt_ann(ann, angle_list, n_annuli, fwhm, radius_int, annulus_width,
                     delta_rot, nframes, verbose):
    """ Optimized median subtraction for a given annulus.
    """
    n = array.shape[0]

    # The annulus is built, and the corresponding PA thresholds for frame
    # rejection are calculated. The PA rejection is calculated at center of
    # the annulus
    pa_thr, inner_radius, _ = _define_annuli(angle_list, ann, n_annuli, fwhm,
                                             radius_int, annulus_width,
                                             delta_rot, 1, verbose)

    indices = get_annulus(array[0], inner_radius, annulus_width,
                          output_indices=True)
    yy = indices[0]
    xx = indices[1]
    matrix = array[:, yy, xx]  # shape [n x npx_annulus]
    matrix_res = np.zeros_like(matrix)

    # A second optimized psf reference is subtracted from each frame.
    # For each frame we find ``nframes``, depending on the PA threshold,
    # to construct this optimized psf reference
    for frame in range(n):
        if pa_thr != 0:
            indices_left = _find_indices(angle_list, frame, pa_thr, nframes)
            matrix_disc = matrix[indices_left]
        else:
            matrix_disc = matrix

        ref_psf_opt = np.median(matrix_disc, axis=0)
        curr_frame = matrix[frame]
        subtracted = curr_frame - ref_psf_opt
        matrix_res[frame] = subtracted

    return matrix_res, yy, xx





#! /usr/bin/env python

"""
Module with simplex (Nelder-Mead) optimization for defining the flux and 
position of a companion using the Negative Fake Companion.
"""
from __future__ import print_function

import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from .simplex_fmerit import chisquare
from ..var import frame_center
from ..conf import time_ini, timing
from ..conf.utils_conf import sep


__all__ = ['firstguess']



def firstguess_from_coord(planet, center, cube, angs, PLSC, psf, 
                          fwhm, annulus_width, aperture_radius, ncomp, 
                          cube_ref=None, svd_mode='lapack', scaling=None, 
                          fmerit='sum', imlib='opencv', interpolation='lanczos4',
                          collapse='median', f_range=None, display=False,
                          verbose=True, save=False, **kwargs):
    """
    Determine a first guess for the flux of a companion at a given position 
    in the cube by doing a simple grid search evaluating the reduced chi2.
    
    Parameters
    ----------
    planet: numpy.array
        The (x,y) position of the planet in the pca processed cube.
    center: numpy.array
        The (x,y) position of the cube center.
    cube: numpy.array
        The cube of fits images expressed as a numpy.array. 
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array.         
    PLSC: float
        The platescale, in arcsec per pixel.
    psf: numpy.array
        The scaled psf expressed as a numpy.array. 
    fwhm : float
        The FHWM in pixels.           
    annulus_width: int, optional
        The width in terms of the FWHM of the annulus on which the PCA is done.       
    aperture_radius: int, optional
        The radius of the circular aperture in terms of the FWHM.
    ncomp: int
        The number of principal components. 
    cube_ref : array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
    scaling : {'temp-mean', 'temp-standard'} or None, optional
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done. 
    fmerit : {'sum', 'stddev'}, string optional
        Chooses the figure of merit to be used. stddev works better for close in
        companions sitting on top of speckle noise.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is used when measuring the function of
        merit (instead of a single final frame).
    f_range: numpy.array, optional
        The range of flux tested values. If None, 20 values between 0 and 5000
        are tested.
    display: boolean, optional
        If True, the figure chi2 vs. flux is displayed.
    verbose: boolean
        If True, display intermediate info in the shell.
    save: boolean, optional
        If True, the figure chi2 vs. flux is saved.
    kwargs: dict, optional
        Additional parameters are passed to the matplotlib plot method.
        
    Returns
    -------
    out : numpy.array
        The radial coordinates and the flux of the companion.
                
    """  
    xy = planet-center
    r0 = np.sqrt(xy[0]**2+xy[1]**2)
    theta0 = np.mod(np.arctan2(xy[1],xy[0])/np.pi*180,360) 

    if f_range is not None:    
        n = f_range.shape[0]
    else:
        n = 20
        f_range = np.linspace(0,5000,n)
    
    chi2r = []
    if verbose:
        print('Step | flux    | chi2r')
        
    counter = 0
    for j, f_guess in enumerate(f_range):
        chi2r.append(chisquare((r0,theta0,f_guess), cube, angs, PLSC, psf, 
                                fwhm, annulus_width, aperture_radius,(r0,theta0),
                                ncomp, cube_ref, svd_mode, scaling, fmerit,
                                collapse, imlib, interpolation))
        if chi2r[j] > chi2r[j-1]:
            counter += 1
        if counter == 4:
            break
        if verbose:
            print('{}/{}   {:.3f}   {:.3f}'.format(j+1,n,f_guess,chi2r[j]))

    chi2r = np.array(chi2r)
    f0 = f_range[chi2r.argmin()]  

    if display:
        plt.figure(figsize=kwargs.pop('figsize',(8,4)))
        plt.title(kwargs.pop('title',''))
        plt.xlim(f_range[0], f_range[:chi2r.shape[0]].max())
        plt.ylim(chi2r.min()*0.9, chi2r.max()*1.1)
        plt.plot(f_range[:chi2r.shape[0]],chi2r,
                 linestyle=kwargs.pop('linestyle','-'),
                 color=kwargs.pop('color','gray'),
                 marker=kwargs.pop('marker','.'),
                 markerfacecolor='r', markeredgecolor='r', **kwargs)
        plt.xlabel('flux')
        plt.ylabel(r'$\chi^2_{r}$')
        plt.grid('on')
    if save:
        plt.savefig('chi2rVSflux.pdf')
    if display:
        plt.show()

    return (r0,theta0,f0)       


def firstguess_simplex(p, cube, angs, psf, plsc, ncomp, fwhm, annulus_width, 
                       aperture_radius, cube_ref=None, svd_mode='lapack', 
                       scaling=None, fmerit='sum', imlib='opencv',
                       interpolation='lanczos4', collapse='median', p_ini=None,
                       options=None, verbose=False, **kwargs):
    """
    Determine the position of a companion using the negative fake companion 
    technique and a standard minimization algorithm (Default=Nelder-Mead) .
    
    Parameters
    ----------
    
    p : np.array
        Estimate of the candidate position.
    cube: numpy.array
        The cube of fits images expressed as a numpy.array. 
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array. 
    psf: numpy.array
        The scaled psf expressed as a numpy.array.        
    plsc: float
        The platescale, in arcsec per pixel.
    ncomp: int
        The number of principal components.  
    fwhm : float
        The FHWM in pixels.   
    annulus_width: int, optional
        The width in terms of the FWHM of the annulus on which the PCA is done.       
    aperture_radius: int, optional
        The radius of the circular aperture in terms of the FWHM.
    cube_ref : array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
    scaling : {'temp-mean', 'temp-standard'} or None, optional
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done. 
    fmerit : {'sum', 'stddev'}, string optional
        Chooses the figure of merit to be used. stddev works better for close in
        companions sitting on top of speckle noise.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is used when measuring the function of
        merit (instead of a single final frame).
    p_ini : np.array
        Position (r, theta) of the circular aperture center.
    options: dict, optional
        The scipy.optimize.minimize options.
    verbose : boolean, optional
        If True, informations are displayed in the shell.
        
    Returns
    -------
    out : scipy.optimize.minimize solution object
        The solution of the minimization algorithm.
        
    """    
    if verbose:
        msg = '\n{} minimization is running...'
        print(msg.format(options.get('method','Nelder-Mead')))
     
    if p_ini is None:
        p_ini = p

    solu = minimize(chisquare, p, args=(cube, angs, plsc, psf, fwhm, annulus_width,
                                        aperture_radius, p_ini, ncomp, cube_ref,
                                        svd_mode, scaling, fmerit, collapse,
                                        imlib, interpolation),
                    method=options.pop('method','Nelder-Mead'),
                    options=options, **kwargs)                       

    if verbose:
        print(solu)
    return solu
    


def firstguess(cube, angs, psfn, ncomp, plsc, planets_xy_coord, fwhm=4, 
               annulus_width=3, aperture_radius=4, cube_ref=None, 
               svd_mode='lapack', scaling=None, fmerit='sum', imlib='opencv',
               interpolation='lanczos4', collapse='median', p_ini=None,
               f_range=None, simplex=True, simplex_options=None, display=False,
               verbose=True, save=False, figure_options=None):
    """ Determines a first guess for the position and the flux of a planet.
        
    We process the cube without injecting any negative fake companion. 
    This leads to the visual detection of the planet(s). For each of them,
    one can estimate the (x,y) coordinates in pixel for the position of the 
    star, as well as the planet(s). 

    From the (x,y) coordinates in pixels for the star and planet(s), we can 
    estimate a preliminary guess for the position and flux for each planet
    by using the method "firstguess_from_coord". The argument "f_range" allows
    to indicate prior limits for the flux (optional, default: None). 
    This step can be reiterate to refine the preliminary guess for the flux.

    We can go a step further by using a Simplex Nelder_Mead minimization to
    estimate the first guess based on the preliminary guess.
           
    Parameters
    ----------
    cube: numpy.array
        The cube of fits images expressed as a numpy.array. 
    angs: numpy.array
        The parallactic angle fits image expressed as a numpy.array.  
    psfn: numpy.array
        The centered and normalized (flux in a 1*FWHM aperture must equal 1) 
        PSF 2d-array.
    ncomp: int
        The number of principal components.         
    plsc: float
        The platescale, in arcsec per pixel.  
    planet_xy_coord: array or list
        The list of (x,y) positions of the planets.
    fwhm : float, optional
        The FHWM in pixels.
    annulus_width: int, optional
        The width in terms of the FWHM of the annulus on which the PCA is done.       
    aperture_radius: int, optional
        The radius of the circular aperture in terms of the FWHM.
    cube_ref : array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    svd_mode : {'lapack', 'randsvd', 'eigen', 'arpack'}, str optional
        Switch for different ways of computing the SVD and selected PCs.
    scaling : {'temp-mean', 'temp-standard'} or None, optional
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done and with 
        "temp-standard" temporal mean centering plus scaling to unit variance 
        is done. 
    fmerit : {'sum', 'stddev'}, string optional
        Chooses the figure of merit to be used. stddev works better for close in
        companions sitting on top of speckle noise.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean', None}, str or None, optional
        Sets the way of collapsing the frames for producing a final image. If
        None then the cube of residuals is used when measuring the function of
        merit (instead of a single final frame).
    p_ini: numpy.array
        Position (r, theta) of the circular aperture center.        
    f_range: numpy.array, optional
        The range of flux tested values. If None, 20 values between 0 and 5000
        are tested.
    simplex: boolean, optional
        If True, the Nelder-Mead minimization is performed after the flux grid
        search.
    simplex_options: dict, optional
        The scipy.optimize.minimize options.
    display: boolean, optional
        If True, the figure chi2 vs. flux is displayed.
    verbose: boolean
        If True, display intermediate info in the shell.
    save: boolean, optional
        If True, the figure chi2 vs. flux is saved.
    figure_options: dict, optional
        Additional parameters are passed to the matplotlib plot method.    

    Returns
    -------
    out : The radial coordinates and the flux of the companion.

    Notes
    -----
    Polar angle is not the conventional NORTH-TO-EAST P.A.
    """
    if verbose:  start_time = time_ini()

    if figure_options is None:
        figure_options = {'color':'gray', 'marker':'.', 
                          'title':r'$\chi^2_{r}$ vs flux'}
        
    planets_xy_coord = np.array(planets_xy_coord)
    n_planet = planets_xy_coord.shape[0]

    center_xy_coord = np.array(frame_center(cube[0]))

    if f_range is None:  
        f_range = np.linspace(0,5000,20)
    if simplex_options is None:  
        simplex_options = {'xtol':1e-1, 'maxiter':500, 'maxfev':1000}

    r_0 = np.zeros(n_planet)
    theta_0 = np.zeros_like(r_0)
    f_0 = np.zeros_like(r_0)
    
    for index_planet in range(n_planet):    
        if verbose:
            print('\n'+sep)
            print('             Planet {}           '.format(index_planet))
            print(sep+'\n')
            msg2 = 'Planet {}: flux estimation at the position [{},{}], running ...'
            print(msg2.format(index_planet,planets_xy_coord[index_planet,0],
                              planets_xy_coord[index_planet,1]))
        
        res_init = firstguess_from_coord(planets_xy_coord[index_planet],
                                         center_xy_coord, cube, angs, plsc, psfn,
                                         fwhm, annulus_width, aperture_radius,
                                         ncomp, f_range=f_range,
                                         cube_ref=cube_ref, svd_mode=svd_mode,
                                         scaling=scaling, fmerit=fmerit,
                                         imlib=imlib,
                                         interpolation=interpolation,
                                         collapse=collapse, display=display,
                                         verbose=verbose, save=save,
                                         **figure_options)
        r_pre, theta_pre, f_pre = res_init
                                                                                                                    
        if verbose:
            msg3 = 'Planet {}: preliminary guess: (r, theta, f)=({:.1f}, {:.1f}, {:.1f})'
            print(msg3.format(index_planet,r_pre, theta_pre, f_pre))
        
        if simplex:
            if verbose:
                msg4 = 'Planet {}: Simplex Nelder-Mead minimization, running ...'
                print(msg4.format(index_planet))
                                                         
            res = firstguess_simplex((r_pre,theta_pre,f_pre), cube, angs, psfn,
                                     plsc, ncomp, fwhm, annulus_width, 
                                     aperture_radius, cube_ref=cube_ref, 
                                     svd_mode=svd_mode, scaling=scaling,
                                     fmerit=fmerit, imlib=imlib,
                                     interpolation=interpolation,
                                     collapse=collapse, p_ini=p_ini,
                                     options=simplex_options, verbose=False)
            
            r_0[index_planet], theta_0[index_planet], f_0[index_planet] = res.x
            if verbose:
                msg5 = 'Planet {}: Success: {}, nit: {}, nfev: {}, chi2r: {}'
                print(msg5.format(index_planet,res.success,res.nit,res.nfev, 
                                  res.fun))
                print('message: {}'.format(res.message))
            
        else:
            if verbose:
                msg4bis = 'Planet {}: Simplex Nelder-Mead minimization skipped.'
                print(msg4bis.format(index_planet))            
            r_0[index_planet] = r_pre
            theta_0[index_planet] = theta_pre
            f_0[index_planet] = f_pre                               

        if verbose:            
            centy, centx = frame_center(cube[0])
            posy = r_0 * np.sin(np.deg2rad(theta_0[index_planet])) + centy
            posx = r_0 * np.cos(np.deg2rad(theta_0[index_planet])) + centx
            msg6 = 'Planet {}: simplex result: (r, theta, f)=({:.3f}, {:.3f}'
            msg6 += ', {:.3f}) at \n          (X,Y)=({:.2f}, {:.2f})'
            print(msg6.format(index_planet, r_0[index_planet],
                              theta_0[index_planet], f_0[index_planet], posx[0], posy[0]))
    
    if verbose:
        print('\n', sep, '\nDONE !\n', sep)
        timing(start_time)

    return r_0, theta_0, f_0




from __future__ import (absolute_import)

from . import preproc
from . import conf
from . import fits
from . import frdiff
from . import leastsq
from . import llsg
from . import madi
from . import negfc
from . import nmf
from . import pca
from . import phot
from . import stats
from . import var

__version__ = "0.9.3"







#! /usr/bin/env python

"""
Module for stats of a fits-cube.
"""

from __future__ import division 

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['cube_stats_aperture',
           'cube_stats_annulus']

import numpy as np
import scipy.stats
from matplotlib import pyplot as plt
from ..var import get_annulus, get_circle, get_annulus_cube


def cube_stats_aperture(arr, radius, xy=None, plot=False, full_output=False):                                                 
    """Calculates statistics in an aperture on a 2D or 3D array and plots the 
    variation of the mean, median and standard deviation as a functions of time.
    
    Parameters
    ----------
    arr : array_like
        Input array.
    radius : int
        Radius.
    xy : tuple of ints
        Corrdinated of the center of the aperture
    plot : None,1,2, optional
        If 1 or True it plots the mean, std_dev and max. Also the histogram. 
        If 2 it also plots the linear correlation between the median and the
        std_dev. 
    full_output : {False,True}, optional
        If true it returns mean, std_dev, median, if false just the mean.
        
    Returns
    -------
    If full_out is true it returns the sum, mean, std_dev, median. If false 
    only the mean.
    """
    if arr.ndim == 2:        
        if xy is not None:
            x, y = xy
            circle = get_circle(arr, radius, output_values=True, cy=y, cx=x)
        else:
            circle = get_circle(arr, radius, output_values=True)
        maxi = circle.max()
        mean = circle.mean()
        std_dev = circle.std()
        median = np.median(circle)
            
    if arr.ndim == 3:
        n = arr.shape[0]
        mean = np.empty(n)
        std_dev = np.empty(n)
        median = np.empty(n)
        maxi = np.empty(n)
        
        values_circle = []
        for i in range(n):
            if xy is not None:
                x, y = xy
                circle = get_circle(arr[i], radius, output_values=True, cy=y, cx=x)
            else:
                circle = get_circle(arr[i], radius, output_values=True)
            values_circle.append(circle)
            maxi[i] = circle.max()
            mean[i] = circle.mean()
            std_dev[i] = circle.std()
            median[i] = np.median(circle)
        values_circle = np.array(values_circle).flatten()
        
        if plot==1 or plot==2:
            plt.figure('Image crop (first slice)', figsize=(12,3))
            if xy is not None:
                x, y = xy
                temp = get_circle(arr[0], radius, cy=y, cx=x)
            else:
                temp = get_circle(arr[0], radius)
            ax1 = plt.subplot(1, 2, 1)
            ax1.imshow(arr[0], origin = 'lower', interpolation="nearest", 
                       cmap = plt.get_cmap('gray'), alpha=0.8)
            ax1.imshow(temp, origin = 'lower', interpolation="nearest", 
                       cmap = plt.get_cmap('CMRmap'), alpha=0.6)                     
            plt.axis('on')
            ax2 = plt.subplot(1, 2, 2)
            ax2.hist(values_circle, bins=int(np.sqrt(values_circle.shape[0])),
                     alpha=0.5, histtype='stepfilled', label='Histogram')
            ax2.legend()
            ax2.tick_params(axis='x', labelsize=8)

            fig = plt.figure('Stats in annulus', figsize=(12, 6))
            fig.subplots_adjust(hspace=0.15)
            ax1 = plt.subplot(3, 1, 1)
            std_of_means = np.std(mean)
            median_of_means = np.median(mean)
            lab = 'mean (median={:.1f}, stddev={:.1f})'.format(median_of_means,
                                                               std_of_means)
            ax1.axhline(median_of_means, alpha=0.5, color='gray', lw=2, ls='--')
            ax1.plot(mean, '.-', label=lab, lw = 0.8, alpha=0.6, marker='o', 
                     color='b')
            ax1.legend(loc='best', fancybox=True).get_frame().set_alpha(0.5)
            ax1.grid(True)
            plt.setp(ax1.get_xticklabels(), visible=False)
            
            ax2 = plt.subplot(3, 1, 2, sharex=ax1)
            ax2.plot(std_dev, '.-', label='std_dev', lw = 0.8, alpha=0.6, 
                     marker='o', color='r')
            ax2.legend(loc='best', fancybox=True).get_frame().set_alpha(0.5)
            ax2.grid(True)
            plt.setp(ax2.get_xticklabels(), visible=False)

            ax3 = plt.subplot(3, 1, 3, sharex=ax1)
            ax3.plot(maxi, '.-', label='max', lw=0.8, alpha=0.6, marker='o',
                     color='g')
            ax3.legend(loc='best', fancybox=True).get_frame().set_alpha(0.5)  
            ax3.grid(True)  

            if plot==2:
                plt.figure('Std_dev - mean in annulus', figsize=(4, 4))                
                plt.scatter(std_dev, mean, alpha=0.6)
                m, b = np.polyfit(std_dev, mean, 1)
                corr, _ = scipy.stats.pearsonr(mean, std_dev)
                plt.plot(std_dev, m*std_dev + b, '-', label=corr, alpha=0.6)
                plt.xlabel('Mean')
                plt.ylabel('Standard deviation')
                plt.legend()
    
    if full_output:
        return mean, std_dev, median, maxi
    else:
        return mean


def cube_stats_annulus(array, inner_radius, size, plot=None, full_out=False):
    """Calculates statistics in a centered annulus on a 2D or 3D array and 
    plots the variation of the mean, median and standard deviation as a 
    functions of time.
    
    Parameters
    ----------
    array : array_like
        Input array.
    inner_radius : int
        Annulus inner radius.
    size : int
        How many pixels in radial direction contains the annulus.
    plot : None,1,2, optional
        If 1 or True it plots the mean, std_dev and max. Also the histogram. 
        If 2 it also plots the linear correlation between the median and the
        std_dev. 
    full_out : {False,True}, optional
        If true it returns mean, std_dev, median, if false just the mean.
        
    Returns
    -------
    If full_out is true it returns mean, std_dev, median, if false 
    only the mean.
    """
    if array.ndim==2:
        arr = array.copy()    
        
        annulus = get_annulus(arr, inner_radius, size, output_values=True)
        mean = annulus.mean()
        std_dev = annulus.std()
        median = np.median(annulus)
        maxi = annulus.max()
            
    if array.ndim==3:
        n = array.shape[0]
        mean = np.empty(n)
        std_dev = np.empty(n)
        median = np.empty(n)
        maxi = np.empty(n)
        
        for i in range(n):
            arr = array[i].copy() 
            annulus = get_annulus(arr, inner_radius, size, output_values=True)
            mean[i] = annulus.mean()
            std_dev[i] = annulus.std()
            median[i] = np.median(annulus)
            maxi[i] = annulus.max()
            
        if plot==1 or plot==2:
            plt.figure('Image crop (first slice)', figsize=(12,3))
            temp = get_annulus_cube(array, inner_radius, size)
            ax1 = plt.subplot(1, 2, 1)
            ax1.imshow(array[0], origin = 'lower', interpolation="nearest", 
                       cmap = plt.get_cmap('gray'), alpha=0.8) 
            ax1.imshow(temp[0], origin = 'lower', interpolation="nearest", 
                       cmap = plt.get_cmap('CMRmap'), alpha=0.6)                           
            plt.axis('on')
            ax2 = plt.subplot(1, 2, 2)
            values = temp[np.where(temp>0)]
            ax2.hist(values.ravel(), bins=int(np.sqrt(values.shape[0])),
                     alpha=0.5, histtype='stepfilled', label='Histogram')
            ax2.legend()
            ax2.tick_params(axis='x', labelsize=8)
            
            fig = plt.figure('Stats in annulus', figsize=(12, 6))
            fig.subplots_adjust(hspace=0.15)
            ax1 = plt.subplot(3, 1, 1)
            std_of_means = np.std(mean)
            median_of_means = np.median(mean)
            lab = 'mean (median={:.1f}, stddev={:.1f})'.format(median_of_means,
                                                               std_of_means)
            ax1.axhline(median_of_means, alpha=0.5, color='gray', lw=2, ls='--')
            ax1.plot(mean, '.-', label=lab, lw = 0.8, alpha=0.6, marker='o', 
                     color='b')
            ax1.legend(loc='best', fancybox=True).get_frame().set_alpha(0.5)
            ax1.grid(True)
            plt.setp(ax1.get_xticklabels(), visible=False)
            
            ax2 = plt.subplot(3, 1, 2, sharex=ax1)
            ax2.plot(std_dev, '.-', label='std_dev', lw = 0.8, alpha=0.6, 
                     marker='o', color='r')
            ax2.legend(loc='best', fancybox=True).get_frame().set_alpha(0.5)
            ax2.grid(True)
            plt.setp(ax2.get_xticklabels(), visible=False)

            ax3 = plt.subplot(3, 1, 3, sharex=ax1)
            ax3.plot(maxi, '.-', label='max', lw=0.8, alpha=0.6, marker='o',
                     color='g')
            ax3.legend(loc='best', fancybox=True).get_frame().set_alpha(0.5)  
            ax3.grid(True)          
            
            if plot==2:
                plt.figure('Std_dev - mean in annulus', figsize=(4, 4))                
                plt.scatter(std_dev, mean, alpha=0.6)
                m, b = np.polyfit(std_dev, mean, 1)
                corr, _ = scipy.stats.pearsonr(mean, std_dev)
                plt.plot(std_dev, m*std_dev + b, '-', label=corr, alpha=0.6)
                plt.xlabel('Mean')
                plt.ylabel('Standard deviation')
                plt.legend()

    if full_out:
        return mean, std_dev, median, maxi
    else:
        return mean









#! /usr/bin/env python

"""
Module with S/N calculation functions.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez, O. Absil @ ULg'
__all__ = ['snr_ss',
           'snr_peakstddev',
           'snrmap',
           'snrmap_fast']

import numpy as np
import itertools as itt
import photutils
from skimage import draw
from matplotlib import pyplot as plt
from astropy.convolution import convolve, Tophat2DKernel
from astropy.stats import median_absolute_deviation as mad
from multiprocessing import Pool, cpu_count
from ..conf.utils_conf import eval_func_tuple
from ..conf import time_ini, timing
from ..var import get_annulus, frame_center, dist, pp_subplots


def snrmap(array, fwhm, plot=False, mode='sss', source_mask=None, nproc=None,
           save_plot=None, plot_title=None):
    """Parallel implementation of the SNR map generation function. Applies the 
    S/N function (small samples penalty) at each pixel.
    
    Parameters
    ----------
    array : array_like
        Input frame.
    fwhm : float
        Size in pixels of the FWHM.
    plot : {False, True}, bool optional
        If True plots the SNR map. 
    mode : {'sss', 'peakstddev'}, string optional
        'sss' uses the approach with the small sample statistics penalty and
        'peakstddev' uses the peak(aperture)/std(annulus) version.
    source_mask : array_like, optional
        If exists, it takes into account existing sources. The mask is a ones
        2d array, with the same size as the input frame. The centers of the 
        known sources have a zero value.
    nproc : int or None
        Number of processes for parallel computing.
    save_plot: string
        If provided, the snr map is saved to this path
    plot_title: string
        If provided, the snr map is titled
    
    Returns
    -------
    snrmap : array_like
        Frame with the same size as the input frame with each pixel.
        
    """
    start_time = time_ini()
    if not array.ndim==2:
        raise TypeError('Input array is not a 2d array or image.')
    if plot:  plt.close('snr')
        
    sizey, sizex = array.shape
    snrmap = np.zeros_like(array)
    width = min(sizey,sizex)/2 - 1.5*fwhm
    mask = get_annulus(array, (fwhm/2)+1, width)
    mask = np.ma.make_mask(mask)
    yy, xx = np.where(mask)
    coords = zip(xx,yy)
        
    if not nproc:  
        nproc = int((cpu_count()/2))  # Hyper-threading doubles the # of cores
    
    if mode == 'sss':
        F = snr_ss
    elif mode == 'peakstddev':
        F = snr_peakstddev
    else:
        raise TypeError('\nMode not recognized.')
    
    if source_mask is None:
        pool = Pool(processes=int(nproc))                                        
        res = pool.map(eval_func_tuple, zip(itt.repeat(F),itt.repeat(array),
                                            coords, itt.repeat(fwhm),
                                            itt.repeat(True)))
        res = np.array(res)
        pool.close()
        yy = res[:, 0]
        xx = res[:, 1]
        snr = res[:, 2]
        snrmap[yy.astype('int'), xx.astype('int')] = snr
    else:
        # checking the mask with the sources
        if not array.shape == source_mask.shape:
            raise RuntimeError('Source mask has wrong size.')
        if source_mask[source_mask == 0].shape[0] == 0:
            msg = 'Input source mask is empty.'
            raise RuntimeError(msg)
        if source_mask[source_mask == 0].shape[0] > 20:
            msg = 'Input source mask is too crowded. Check its validity.'
            raise RuntimeError(msg)
        
        soury, sourx = np.where(source_mask == 0)
        sources = []
        ciry = []; cirx = []; anny = []; annx = []
        array_sources = array.copy()
        centery, centerx = frame_center(array)
        for (y,x) in zip(soury,sourx):
            radd = dist(centery, centerx, y, x)
            if int(np.floor(radd)) < centery - np.ceil(fwhm):
                sources.append((y,x))
        
        for source in sources:
            y, x = source        
            radd = dist(centery, centerx, y, x)
            tempay, tempax = get_annulus(array, int(np.floor(radd-fwhm)), 
                                    int(np.ceil(2*fwhm)), output_indices=True)
            tempcy, tempcx = draw.circle(y, x, int(np.ceil(1*fwhm)))
            # masking the source position (using the MAD of pixels in annulus)
            array_sources[tempcy, tempcx] = mad(array[tempay, tempax])
            ciry += list(tempcy); cirx += list(tempcx)
            anny += list(tempay); annx += list(tempax)

        # coordinates of annulus without the sources
        coor_ann = [(y,x) for (y,x) in zip(anny, annx) if (y,x) not in zip(ciry, cirx)]

        # coordinates of the rest of the frame without the annulus
        coor_rest = [(y,x) for (y,x) in zip(yy, xx) if (y,x) not in coor_ann]
        
        pool1 = Pool(processes=int(nproc))
        res = pool1.map(eval_func_tuple, zip(itt.repeat(F),itt.repeat(array),
                                             coor_rest, itt.repeat(fwhm),
                                             itt.repeat(True)))
        res = np.array(res)
        pool1.close()
        yy = res[:,0]
        xx = res[:,1]
        snr = res[:,2]
        snrmap[yy.astype('int'), xx.astype('int')] = snr
        
        pool2 = Pool(processes=int(nproc))
        res = pool2.map(eval_func_tuple, zip(itt.repeat(F),
                                             itt.repeat(array_sources),
                                             coor_ann, itt.repeat(fwhm),
                                             itt.repeat(True)))
        res = np.array(res)
        pool2.close()
        yy = res[:,0]
        xx = res[:,1]
        snr = res[:,2]
        snrmap[yy.astype('int'), xx.astype('int')] = snr
    
    if plot:  
        pp_subplots(snrmap, colorb=True, title='S/N map')

    # Option to save snrmap in angular scale, using Keck NIRC2's ~0.01 pixel scale
    # In this case, set plot = False
    elif save_plot is not None:
        pp_subplots(snrmap, colorb=True, title=plot_title, save=save_plot,
                    vmin=-1, vmax=5, angscale=True, getfig = True)
        
    print("S/N map created using {:} processes.".format(nproc))
    timing(start_time)
    return snrmap
   
   
def snrmap_fast(array, fwhm, nproc=None, plot=False, verbose=True):
    """ Serial implementation of the S/N map generation function. To be used as
    a quick proxy of the snrmap generated using the small samples statistics
    definition. 
    
    Parameters
    ----------
    array : array_like
        Input frame.
    fwhm : float
        Size in pixels of the FWHM.
    nproc : int or None
        Number of processes for parallel computing.
    plot : {False, True}, bool optional
        If True plots the SNR map. 
    verbose: {True, False}
        Chooses whether to print results or not. 
    
    Returns
    -------
    snrmap : array_like
        Frame with the same size as the input frame with each pixel.
        
    """       
    if verbose:
        start_time = time_ini()
    if not array.ndim == 2:
        raise TypeError('Input array is not a 2d array or image.')
    
    cy, cx = frame_center(array)
    tophat_kernel = Tophat2DKernel(fwhm/2.)
    array = convolve(array, tophat_kernel)
            
    sizey, sizex = array.shape
    snrmap = np.zeros_like(array)
    width = min(sizey,sizex)/2 - 1.5*fwhm    
    mask = get_annulus(array, (fwhm/2)+1, width-1)
    mask = np.ma.make_mask(mask)
    yy, xx = np.where(mask)
    coords = [(x, y) for (x, y) in zip(xx, yy)]
    
    if nproc is None:  
        nproc = int((cpu_count()/2))  # Hyper-threading doubles the # of cores
    
    if nproc == 1:
        for y,x in zip(yy, xx):
            snrmap[y,x] = _snr_approx(array, (x, y), fwhm, cy, cx)[2]
    elif nproc > 1:
        pool = Pool(processes=int(nproc))                                        
        res = pool.map(eval_func_tuple, zip(itt.repeat(_snr_approx),
                                            itt.repeat(array), coords,
                                            itt.repeat(fwhm), itt.repeat(cy),
                                            itt.repeat(cx)))
        res = np.array(res)
        pool.close()
        yy = res[:, 0]
        xx = res[:, 1]
        snr = res[:, 2]
        snrmap[yy.astype('int'), xx.astype('int')] = snr
        
    if plot:
        pp_subplots(snrmap, colorb=True, title='SNRmap')
     
    if verbose:    
        print("S/N map created using {:} processes.".format(nproc))
        timing(start_time)
    return snrmap


def _snr_approx(array, source_xy, fwhm, centery, centerx):
    """
    array - frame convolved with top hat kernel 
    """
    sourcex, sourcey = source_xy
    rad = dist(centery, centerx, sourcey, sourcex)
    ind_aper = draw.circle(sourcey, sourcex, fwhm/2.)
    # noise : STDDEV in convolved array of 1px wide annulus (while 
    # masking the flux aperture) * correction of # of resolution elements
    ind_ann = draw.circle_perimeter(int(centery), int(centerx), int(rad))
    array2 = array.copy()
    array2[ind_aper] = array[ind_ann].mean()   # quick-n-dirty mask
    n2 = ((2*np.pi*rad)/fwhm) - 1
    noise = array2[ind_ann].std()*np.sqrt(1+(1/n2))
    # signal : central px minus the mean of the pxs (masked) in 1px annulus
    signal = array[sourcey, sourcex] - array2[ind_ann].mean()
    snr = signal / noise
    return sourcey, sourcex, snr
    

def snr_ss(array, source_xy, fwhm, out_coor=False, plot=False, verbose=False,
           full_output=False):
    # Leave the order of parameters as it is, the same for both snr functions
    # to be compatible with the snrmap parallel implementation
    """Calculates the SNR (signal to noise ratio) of a single planet in a 
    post-processed (e.g. by LOCI or PCA) frame. Uses the approach described in 
    Mawet et al. 2014 on small sample statistics, where a student t-test (eq. 9)
    can be used to determine S/N (and contrast) in high contrast imaging.
    
    Parameters
    ----------
    array : array_like, 2d
        Post-processed frame where we want to measure SNR.
    source_xy : tuple of floats
        X and Y coordinates of the planet or test speckle.
    fwhm : float
        Size in pixels of the FWHM.
    out_coor: {False, True}, bool optional
        If True returns back the snr value and the y, x input coordinates. In 
        this case it overrides the full_output parameter.
    plot : {False, True}, bool optional
        Plots the frame and the apertures considered for clarity. 
    verbose: {True, False}, bool optional
        Chooses whether to print some output or not. 
    full_output: {False, True}, bool optional
        If True returns back the snr value, the y, x input coordinates, noise 
        and flux.   
    
    Returns
    -------
    snr : float
        Value of the SNR for the given planet or test speckle.
        
    sourcey, sourcex, f_source, fluxes.std(), snr
    
    """
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array')
    if out_coor and full_output:
        raise TypeError('One of the 2 must be False')
    
    sourcex, sourcey = source_xy 
    
    centery, centerx = frame_center(array)
    sep = dist(centery,centerx,sourcey,sourcex)
    
    sens = 'clock' #counterclock
        
    angle = np.arcsin(fwhm/2./sep)*2
    number_apertures = int(np.floor(2*np.pi/angle))
    yy = np.zeros((number_apertures))
    xx = np.zeros((number_apertures))
    cosangle = np.cos(angle)
    sinangle = np.sin(angle)
    xx[0] = sourcex - centerx
    yy[0] = sourcey - centery
    for i in range(number_apertures-1):
        if sens == 'clock':
            xx[i+1] = cosangle*xx[i] + sinangle*yy[i] 
            yy[i+1] = cosangle*yy[i] - sinangle*xx[i] 
        elif sens == 'counterclock':
            xx[i+1] = cosangle*xx[i] - sinangle*yy[i] 
            yy[i+1] = cosangle*yy[i] + sinangle*xx[i]           
            
    xx[:] += centerx
    yy[:] += centery 
    rad = fwhm/2.
    apertures = photutils.CircularAperture((xx, yy), r=rad)  # Coordinates (X,Y)
    fluxes = photutils.aperture_photometry(array, apertures, method='exact')

    fluxes = np.array(fluxes['aperture_sum'])
    f_source = fluxes[0].copy()
    fluxes = fluxes[1:]
    n2 = fluxes.shape[0]
    snr = (f_source - fluxes.mean())/(fluxes.std()*np.sqrt(1+(1/n2)))
    
    if verbose:
        msg1 = 'S/N for the given pixel = {:.3f}'
        msg2 = 'Integrated flux in FWHM test aperture = {:.3f}'
        msg3 = 'Mean of background apertures integrated fluxes = {:.3f}'
        msg4 = 'Std-dev of background apertures integrated fluxes = {:.3f}'
        print(msg1.format(snr))
        print(msg2.format(f_source))
        print(msg3.format(fluxes.mean()))
        print(msg4.format(fluxes.std()))
    
    if plot:
        _, ax = plt.subplots(figsize=(6,6))
        ax.imshow(array, origin='lower', interpolation='nearest', alpha=0.5, cmap='gray')
        for i in range(xx.shape[0]):
            # Circle takes coordinates as (X,Y)
            aper = plt.Circle((xx[i], yy[i]), radius=fwhm/2., color='r', 
                              fill=False, alpha=0.8)                                       
            ax.add_patch(aper)
            cent = plt.Circle((xx[i], yy[i]), radius=0.8, color='r', fill=True,
                              alpha=0.5)
            ax.add_patch(cent)
            aper_source = plt.Circle((sourcex, sourcey), radius=0.7, 
                                     color='b', fill=True, alpha=0.5)                                       
            ax.add_patch(aper_source)
        ax.grid('off')
        plt.show()
    
    if out_coor:
        return sourcey, sourcex, snr
    if full_output:
        return sourcey, sourcex, f_source, fluxes.std(), snr
    else:
        return snr
    

def snr_peakstddev(array, source_xy, fwhm, out_coor=False, plot=False,
                   verbose=False):
    """Calculates the S/N (signal to noise ratio) of a single planet in a
    post-processed (e.g. by LOCI or PCA) frame. The signal is taken as the ratio 
    of pixel value of the planet (test speckle) and the noise computed as the 
    standard deviation of the pixels in an annulus at the same radial distance 
    from the center of the frame. The diameter of the signal aperture and the 
    annulus width is in both cases 1 FWHM ~ 1 lambda/D.
    
    Parameters
    ----------
    array : array_like, 2d
        Post-processed frame where we want to measure SNR.
    source_xy : tuple of floats
        X and Y coordinates of the planet or test speckle.
    fwhm : float
        Size in pixels of the FWHM.
    out_coor: {False, True}, bool optional
        If True returns back the snr value and the y, x input coordinates.
    plot : {False, True}, optional
        Plots the frame and the apertures considered for clarity. 
    verbose: {True, False}
        Chooses whether to print some intermediate results or not.    
        
    Returns
    -------
    snr : float
        Value of the SNR for the given planet or test speckle.
    
    """     
    sourcex, sourcey = source_xy
    centery, centerx = frame_center(array)
    rad = dist(centery,centerx,sourcey,sourcex)  
    
    array = array + np.abs(array.min()) 
    inner_rad = np.round(rad)-(fwhm/2.)
    an_coor = get_annulus(array, inner_rad, fwhm, output_indices=True)
    ap_coor = draw.circle(sourcey, sourcex, int(np.ceil(fwhm/2.)))
    array2 = array.copy()
    array2[ap_coor] = array[an_coor].mean()   # we 'mask' the flux aperture
    stddev = array2[an_coor].std()
    peak = array[sourcey, sourcex] 
    snr = peak / stddev
    if verbose:
        msg = "S/N = {:.3f}, Peak px = {:.3f}, Noise = {:.3f}"
        print(msg.format(snr, peak, stddev))
    
    if plot:
        _, ax = plt.subplots(figsize=(6,6))
        ax.imshow(array, origin='lower', interpolation='nearest')
        circ = plt.Circle((centerx, centery), radius=inner_rad, color='r', 
                          fill=False) 
        ax.add_patch(circ)
        circ2 = plt.Circle((centerx, centery), radius=inner_rad+fwhm, color='r', 
                           fill=False) 
        ax.add_patch(circ2)
        aper = plt.Circle((sourcex, sourcey), radius=fwhm/2., color='b', 
                          fill=False)   # Coordinates (X,Y)
        ax.add_patch(aper)
        plt.show()
        plt.gray()
        
    if out_coor:
        return sourcey, sourcex, snr
    else:
        return snr


"""
Subpackage ``exlib`` contains external libraries self contained in VIP:
IUWT from https://github.com/ratt-ru/PyMORESANE/
"""
from __future__ import absolute_import

#from .ds9 import *
from .iuwt import *

__all__ = []



#! /usr/bin/env python

"""
Module with utilities.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = []

import sys


sep = '----------------------------------------------------------------------'

def eval_func_tuple(f_args):
    """ Takes a tuple of a function and args, evaluates and returns result"""
    return f_args[0](*f_args[1:])                       


class redirect_output(object):
    """ Context manager for redirecting stdout/err to files"""
    def __init__(self, stdout='', stderr=''):
        self.stdout = stdout
        self.stderr = stderr

    def __enter__(self):
        self.sys_stdout = sys.stdout
        self.sys_stderr = sys.stderr

        if self.stdout:
            sys.stdout = open(self.stdout, 'w')
        if self.stderr:
            if self.stderr == self.stdout:
                sys.stderr = sys.stdout
            else:
                sys.stderr = open(self.stderr, 'w')

    def __exit__(self, exc_type, exc_value, traceback):
        sys.stdout = self.sys_stdout
        sys.stderr = self.sys_stderr
        
    



"""
Subpackage ``madi`` has the ADI cube basic processing (Marois et al. 2006): median
frame subtraction, and annular mode where ``n`` closest frames taking into account
a PA threshold are median collapsed and subtracted.
"""
from __future__ import absolute_import

from .adi_source import *
from .adi_utils import *


"""
Subpackage ``var`` has helping functions such as:
    - image filtering,
    - shapes extraction (annulus, squares subimages, circular apertures),
    - plotting,
    - 2d fitting (Gaussian, Moffat).
"""
from __future__ import absolute_import

from .filters import *
from .fit_2d import *
from .shapes import *
from .utils_var import *


#! /usr/bin/env python

"""
Module with frame resampling/rescaling functions.
"""
__author__ = 'Carlos Alberto Gomez Gonzalez, V. Christiaens @ ULg'
__all__ = ['frame_px_resampling',
           'cube_px_resampling',
           'frame_rescaling',
           'cube_rescaling',
           'check_scal_vector']

import numpy as np
import warnings
try:
    import cv2
    no_opencv = False
except ImportError:
    msg = "Opencv python bindings are missing."
    warnings.warn(msg, ImportWarning)
    no_opencv = True

from skimage.transform import rescale
from scipy.ndimage.interpolation import geometric_transform
from ..var import frame_center


def frame_px_resampling(array, scale, imlib='opencv', interpolation='lanczos4',
                        scale_y=None, scale_x=None, keep_odd=True,
                        full_output=False):
    """ Resamples the pixels of a frame wrt to the center, changing the size
    of the frame. If ``scale`` < 1 then the frame is downsampled and if
    ``scale`` > 1 then its pixels are upsampled.

    Parameters
    ----------
    array : array_like 
        Input frame, 2d array.
    scale : int or float
        Scale factor for upsampling or downsampling the frame.
    imlib : {'opencv', 'skimage'}, str optional
        Library used for image transformations. Opencv is faster than ndimage or
        skimage.
    interpolation : str, optional
        For 'skimage' library: 'nearneig', bilinear', 'bicuadratic', 'bicubic',
        'biquartic', 'biquintic'. The 'nearneig' interpolation is the fastest
        and the 'biquintic' the slowest. The 'nearneig' is the poorer
        option for interpolation of noisy astronomical images.
        For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
        The 'nearneig' interpolation is the fastest and the 'lanczos4' the
        slowest and accurate. 'lanczos4' is the default.
    scale_y : int or float, opt
        Scale factor for upsampling or downsampling the frame along y only. If 
        provided, it takes priority on scale parameter.
    scale_x : int or float, opt
        Scale factor for upsampling or downsampling the frame along x only. If 
        provided, it takes priority on scale parameter.
    keep_odd: bool, opt
        Will slightly modify the scale factor in order for the final frame size
        to keep y and x sizes odd. This keyword does nothing if the input array
        has even dimensions.
    full_output: bool, opt
        If True, it will also return the scale factor (slightly modified if
        ``keep_odd`` was set to True).

    Returns
    -------
    array_resc : array_like 
        Output resampled frame.
    """
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array.')

    if scale_y is None: scale_y = scale
    if scale_x is None: scale_x = scale

    ny = array.shape[0]
    nx = array.shape[1]

    if keep_odd:
        if ny % 2 != 0:  # original y size is odd?
            new_ny = ny * scale_y  # check final size
            if not new_ny > 0.5:
                raise ValueError(
                    "scale_y is too small; resulting array would be 0 size.")
            # final size is decimal? => make it integer
            if new_ny % 1 > 0.5:
                scale_y = (new_ny + 1 - (new_ny % 1)) / ny
            elif new_ny % 1 > 0:
                scale_y = (new_ny - (new_ny % 1)) / ny
            new_ny = ny * scale_y
            # final size is even?
            if new_ny % 2 == 0:
                if scale_y > 1:  # if upscaling => go to closest odd with even-1
                    scale_y = float(new_ny - 1) / ny
                else:
                    scale_y = float(new_ny + 1) / ny
                    # if downscaling => go to closest odd with even+1 (reversible)

        if nx % 2 != 0:  # original x size is odd?
            new_nx = nx * scale_x  # check final size
            if not new_nx > 0.5:
                raise ValueError(
                    "scale_x is too small; resulting array would be 0 size.")
            # final size is decimal? => make it integer
            if new_nx % 1 > 0.5:
                scale_x = (new_nx + 1 - (new_nx % 1)) / nx
            elif new_nx % 1 > 0:
                scale_x = (new_nx - (new_nx % 1)) / nx
            new_nx = nx * scale_x
            # final size is even?
            if new_nx % 2 == 0:
                if scale_x > 1:  # if upscaling => go to closest odd with even-1
                    scale_x = float(new_nx - 1) / nx
                else:  # if downscaling => go to closest odd with even+1 (reversible)
                    scale_x = float(new_nx + 1) / nx

    if imlib == 'skimage':
        if interpolation == 'nearneig':
            order = 0
        elif interpolation == 'bilinear':
            order = 1
        elif interpolation == 'bicuadratic':
            order = 2
        elif interpolation == 'bicubic':
            order = 3
        elif interpolation == 'biquartic':
            order = 4
        elif interpolation == 'biquintic':
            order = 5
        else:
            raise TypeError('Skimage interpolation method not recognized.')

        min_val = np.min(array)
        im_temp = array - min_val
        max_val = np.max(im_temp)
        im_temp /= max_val

        array_resc = rescale(im_temp, scale=(scale_y, scale_x), order=order)
        array_resc *= max_val
        array_resc += min_val

    elif imlib == 'opencv':
        if no_opencv:
            msg = 'Opencv python bindings cannot be imported. Install opencv or'
            msg += ' set imlib to skimage.'
            raise RuntimeError(msg)

        if interpolation == 'bilinear':
            intp = cv2.INTER_LINEAR
        elif interpolation == 'bicubic':
            intp= cv2.INTER_CUBIC
        elif interpolation == 'nearneig':
            intp = cv2.INTER_NEAREST
        elif interpolation == 'lanczos4':
            intp = cv2.INTER_LANCZOS4
        else:
            raise TypeError('Opencv interpolation method not recognized.')

        array_resc = cv2.resize(array.astype(np.float32), (0,0), fx=scale_x,
                                fy=scale_y, interpolation=intp)

    else:
        raise ValueError('Image transformation library not recognized.')

    array_resc /= (scale_y * scale_x)

    if full_output:
        if scale_y == scale_x:
            scale = scale_y
        else:
            scale = (scale_y, scale_x)
        return array_resc, scale
    else:
        return array_resc



def cube_px_resampling(array, scale, imlib='opencv', interpolation='lanczos4',
                       scale_y=None, scale_x=None):
    """ Wrapper of ``frame_px_resampling`` for resampling the frames of a cube
    with a single scale factor. Useful when we need to upsample (upscaling) or
    downsample (pixel binning) a set of frames, e.g. an ADI cube.
    
    Parameters
    ----------
    array : array_like 
        Input frame, 2d array.
    scale : float
        Scale factor for upsampling or downsampling the frames in the cube.
    imlib : str optional
        See the documentation of the ``vip_hci.preproc.frame_px_resampling``
        function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_px_resampling``
        function.
    scale_y : float
        Scale factor for upsampling or downsampling the frame along y. If 
        provided, it takes priority on scale parameter.
    scale_x : float
        Scale factor for upsampling or downsampling the frame along x. If 
        provided, it takes priority on scale parameter.
        
    Returns
    -------
    array_resc : array_like 
        Output cube with resampled frames.

    Notes
    -----
    Be aware that the interpolation used for the pixel rescaling can create
    negative values at the transition from a background composed mostly of zeros
    to a positive valued patch (e.g. an injected PSF template).

    """
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array.')
    
    if scale_y is None: scale_y = scale
    if scale_x is None: scale_x = scale

    array_resc = []
    for i in range(array.shape[0]):
        array_resc.append(frame_px_resampling(array[i], scale=scale, imlib=imlib,
                                              interpolation=interpolation,
                                              scale_y=scale_y, scale_x=scale_x))

    return np.array(array_resc)


### TODO: Merge with ``frame_px_resampling``?
def frame_rescaling(array, ref_y=None, ref_x=None, scale=1.0, imlib='opencv',
                    interpolation='lanczos4', scale_y=None, scale_x=None):
    """ Function to rescale a frame by a factor ``scale``, wrt a reference point
    which by default is the center of the frame (typically the exact location
    of the star). However, it keeps the same dimensions.
    
    Parameters
    ----------
    array : array_like 
        Input frame, 2d array.
    ref_y, ref_x : float, optional
        Coordinates X,Y  of the point with respect to which the rotation will be
        performed. By default the rescaling is done with respect to the center
        of the frame; central pixel if frame has odd size.
    scale : float
        Scaling factor. If > 1, it will upsample the input array equally along y
        and x by this factor.      
    imlib : {'opencv', 'ndimage'}, str optional
        Library used for image transformations. Opencv is faster than ndimage or
        skimage.
    interpolation : str, optional
        For 'ndimage' library: 'nearneig', bilinear', 'bicuadratic', 'bicubic',
        'biquartic', 'biquintic'. The 'nearneig' interpolation is the fastest
        and the 'biquintic' the slowest. The 'nearneig' is the poorer
        option for interpolation of noisy astronomical images.
        For 'opencv' library: 'nearneig', 'bilinear', 'bicubic', 'lanczos4'.
        The 'nearneig' interpolation is the fastest and the 'lanczos4' the
        slowest and accurate. 'lanczos4' is the default.
    scale_y : float
        Scaling factor only for y axis. If provided, it takes priority on scale
        parameter.
    scale_x : float
        Scaling factor only for x axis. If provided, it takes priority on scale
        parameter.

    Returns
    -------
    array_out : array_like
        Resulting frame.
    """
    def _scale_func(output_coords,ref_y=0,ref_x=0, scaling=1.0, scaling_y=None,
                    scaling_x=None):    
        """
        For each coordinate point in a new scaled image (output_coords), 
        coordinates in the image before the scaling are returned. This scaling
        function is used within geometric_transform which, for each point in the
        output image, will compute the (spline) interpolated value at the
        corresponding frame coordinates before the scaling.
        """
        if scaling_y is None:
            scaling_y = scaling
        if scaling_x is None:
            scaling_x = scaling
        return (ref_y+((output_coords[0]-ref_y)/scaling_y), 
                ref_x+((output_coords[1]-ref_x)/scaling_x))
    #---------------------------------------------------------------------------
    if not array.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array.')

    if not ref_y and not ref_x:
        ref_y, ref_x = frame_center(array)

    if imlib == 'ndimage':
        outshap = array.shape

        if interpolation == 'nearneig':
            order = 0
        elif interpolation == 'bilinear':
            order = 1
        elif interpolation == 'bicuadratic':
            order = 2
        elif interpolation == 'bicubic':
            order = 3
        elif interpolation == 'biquartic':
            order = 4
        elif interpolation == 'biquintic':
            order = 5
        else:
            raise TypeError('Scipy.ndimage interpolation method not recognized.')

        array_out = geometric_transform(array, _scale_func, order=order,
                                        output_shape=outshap,
                                        extra_keywords={'ref_y':ref_y,
                                                        'ref_x':ref_x,
                                                        'scaling':scale,
                                                        'scaling_y':scale_y,
                                                        'scaling_x':scale_x})

    elif imlib == 'opencv':
        if no_opencv:
            msg = 'Opencv python bindings cannot be imported. Install opencv or'
            msg += ' set imlib to skimage.'
            raise RuntimeError(msg)

        if interpolation == 'bilinear':
            intp = cv2.INTER_LINEAR
        elif interpolation == 'bicubic':
            intp= cv2.INTER_CUBIC
        elif interpolation == 'nearneig':
            intp = cv2.INTER_NEAREST
        elif interpolation == 'lanczos4':
            intp = cv2.INTER_LANCZOS4
        else:
            raise TypeError('Opencv interpolation method not recognized.')

        if scale_y is None: scale_y = scale
        if scale_x is None: scale_x = scale
        M = np.array([[scale_x,0,(1.-scale_x)*ref_x],
                      [0,scale_y,(1.-scale_y)*ref_y]])
        array_out = cv2.warpAffine(array.astype(np.float32), M, 
                                   (array.shape[1], array.shape[0]), flags=intp)

    else:
        raise ValueError('Image transformation library not recognized.')

    if scale_y==scale_x:
        array_out /= scale ** 2

    return array_out

    
    
def cube_rescaling(array, scaling_list, ref_y=None, ref_x=None, imlib='opencv',
                   interpolation='lanczos4', scaling_y=None, scaling_x=None):
    """ Function to rescale a cube, frame by frame, by a factor ``scale``, with
    respect to position (``ref_y``, ``ref_x``). It calls ``frame_rescaling``
    function.
    
    Parameters
    ----------
    array : array_like 
        Input 3d array, cube.
    scaling_list : 1D-array
        Scale corresponding to each frame in the cube.
    ref_y, ref_x : float, optional
        Coordinates X,Y  of the point with respect to which the rescaling will be
        performed. By default the rescaling is done with respect to the center 
        of the frames; central pixel if the frames have odd size.
    imlib : str optional
        See the documentation of the ``vip_hci.preproc.frame_rescaling``
        function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rescaling``
        function.
    scaling_y : 1D-array or list
        Scaling factor only for y axis. If provided, it takes priority on 
        scaling_list.
    scaling_x : 1D-array or list
        Scaling factor only for x axis. If provided, it takes priority on 
        scaling_list.
        
    Returns
    -------
    array_sc : array_like
        Resulting cube with rescaled frames.
    array_out : array_like
        Median combined image of the rescaled cube.
    """
    if not array.ndim == 3:
        raise TypeError('Input array is not a cube or 3d array')

    array_sc = np.zeros_like(array)
    if scaling_y is None: scaling_y = [None]*array.shape[0]
    if scaling_x is None: scaling_x = [None]*array.shape[0]
    
    if not ref_y and not ref_x:
        ref_y, ref_x = frame_center(array[0])
    
    for i in range(array.shape[0]):
        array_sc[i] = frame_rescaling(array[i], ref_y=ref_y, ref_x=ref_x, 
                                      scale=scaling_list[i], imlib=imlib,
                                      interpolation=interpolation,
                                      scale_y=scaling_y[i],
                                      scale_x=scaling_x[i])
            
    array_out = np.median(array_sc, axis=0)              
    return array_sc, array_out


def check_scal_vector(scal_vec):
    """
    Function to check if the scaling list has the right format to avoid any bug
    in the pca algorithm, in the case of ifs data.
    Indeed, all scaling factors should be >= 1 (i.e. the scaling should be done
    to match the longest wavelength of the cube)

    Parameter:
    ----------
    scal_vec: array_like, 1d OR list

    Returns:
    --------
    scal_vec: array_like, 1d 
        Vector containing the scaling factors (after correction to comply with
        the condition >= 1)

    """
    correct = False

    if isinstance(scal_vec, list):
        scal_list = scal_vec[:]
        nz = len(scal_list)
        scal_vec = np.zeros(nz)
        for ii in range(nz):
            scal_vec[ii] = scal_list[ii]
        correct = True
    elif isinstance(scal_vec,np.ndarray):
        nz = scal_vec.shape[0]
    else:
        raise TypeError('scal_vec is neither a list or an np.ndarray')

    scal_min = np.amin(scal_vec)

    if scal_min < 1:
        correct = True

    if correct:
        for ii in range(nz):
            scal_vec[ii] = scal_vec[ii]/scal_min

    return scal_vec




#! /usr/bin/env python

"""

"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['frame_quick_report']

import numpy as np
import photutils
from skimage import draw
from .snr import snr_ss
from ..conf.utils_conf import sep


def frame_quick_report(array, fwhm, source_xy=None, verbose=True):
    """ Gets information from given frame: Integrated flux in aperture, SNR of
    central pixel (max or given coordinates), mean SNR in aperture.
    
    Parameters
    ----------
    array : array_like
        2d array or input frame.
    fwhm : float
        Size of the FWHM in pixels.
    source_xy : tuple of floats or list (of tuples of floats)
        X and Y coordinates of the center(s) of the source(s).
    verbose : {True, False}, bool optional
        If True prints to stdout the frame info.
        
    Returns
    -------
    source_xy : tuple of floats or list (of tuples of floats)
        X and Y coordinates of the center(s) of the source(s).
    obj_flux : list of floats
        Integrated flux in aperture.
    snr_centpx : list of floats
        S/N of the ``source_xy`` pixels.
    meansnr_pixels : list of floats
        Mean S/N of pixels in 1xFWHM apertures centered on ``source_xy``.
    """
    if not array.ndim == 2:
        raise TypeError('Array is not 2d.')

    if not isinstance(source_xy, list):
        source_xy = [source_xy]

    obj_flux = []
    meansnr_pixels = []
    snr_centpx = []
    for xy in source_xy:
        if xy is not None:
            x, y = xy
            if verbose:
                print(sep)
                print('Coordinates of chosen px (X,Y) = {:},{:}'.format(x,y))
        else:
            y, x = np.where(array == array.max())
            y = y[0]
            x = x[0]
            if verbose:
                print(sep)
                print('Coordinates of Max px (X,Y) = {:},{:}'.format(x,y))

        # we get integrated flux on aperture with diameter=1FWHM
        aper = photutils.CircularAperture((x, y), r=fwhm/2.)
        obj_flux_i = photutils.aperture_photometry(array, aper, method='exact')
        obj_flux_i = obj_flux_i['aperture_sum'][0]

        # we get the mean and stddev of SNRs on aperture
        yy, xx = draw.circle(y, x, fwhm/2.)
        snr_pixels_i = [snr_ss(array, (x_, y_), fwhm, plot=False, verbose=False)\
                        for y_, x_ in zip(yy, xx)]
        meansnr_i = np.mean(snr_pixels_i)
        stdsnr_i = np.std(snr_pixels_i)
        pxsnr_i = snr_ss(array, (x, y), fwhm, plot=False, verbose=False)

        obj_flux.append(obj_flux_i)
        meansnr_pixels.append(meansnr_i)
        snr_centpx.append(pxsnr_i)

        if verbose:
            msg0 = 'Flux in a centered 1xFWHM circular aperture = {:.3f}'
            print(msg0.format(obj_flux_i))
            print('Central pixel S/N = {:.3f}'.format(pxsnr_i))
            print(sep)
            print('Inside a centered 1xFWHM circular aperture:')
            msg1 = 'Mean S/N (shifting the aperture center) = {:.3f}'
            print(msg1.format(meansnr_i))
            msg2 = 'Max S/N (shifting the aperture center) = {:.3f}'
            print(msg2.format(np.max(snr_pixels_i)))
            msg3 = 'stddev S/N (shifting the aperture center) = {:.3f}'
            print(msg3.format(stdsnr_i))
            print(sep)
    
    return source_xy, obj_flux, snr_centpx, meansnr_pixels





#! /usr/bin/env python

"""
Module with fake companion injection functions.
"""

from __future__ import division
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['create_psf_template',
           'psf_norm',
           'cube_inject_companions',
           'frame_inject_companion']

import numpy as np
import photutils
from ..preproc import cube_crop_frames, frame_shift, frame_crop
from ..var import frame_center, fit_2dgaussian, get_circle


# TODO: Check handling even sized frames
def cube_inject_companions(array, psf_template, angle_list, flevel, plsc,
                           rad_dists, n_branches=1, theta=0, imlib='opencv',
                           interpolation='lanczos4', verbose=True):
    """ Injects fake companions in branches, at given radial distances.

    Parameters
    ----------
    array : array_like
        Input frame or cube.
    psf_template : array_like
        2d array with the normalized psf template. It should have an odd shape.
        It's recommended to run the function psf_norm to get a proper PSF
        template. In the ADI+IFS case it must be a 3d array.
    flevel : float or list
        Factor for controlling the brightness of the fake companions.
    plsc : float
        Value of the plsc in arcsec/px.
    rad_dists : list or array 1d
        Vector of radial distances of fake companions in pixels.
    n_branches : int, optional
        Number of azimutal branches.
    theta : float, optional
        Angle in degrees for rotating the position of the first branch that by
        default is located at zero degrees. Theta counts counterclockwise from
        the positive x axis.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    verbose : {True, False}, bool optional
        If True prints out additional information.

    Returns
    -------
    array_out : array_like
        Output array with the injected fake companions.

    """
    if not (array.ndim == 3 or array.ndim == 4):
        raise ValueError('Array is not a cube, 3d or 4d array')
    if array.ndim == 4:
        if not psf_template.ndim == 3:
            raise ValueError('Psf must be a 3d array')

    # ADI case
    if array.ndim == 3:
        if psf_template.shape[1] % 2 == 0:
            raise ValueError("Only odd-sized PSF is accepted")
        ceny, cenx = frame_center(array[0])
        ceny = int(float(ceny))
        cenx = int(float(cenx))
        rad_dists = np.array(rad_dists)
        if not rad_dists[-1] < array[0].shape[0]/2.:
            msg = 'rad_dists last location is at the border (or outside) '
            msg += 'of the field'
            raise ValueError(msg)

        size_fc = psf_template.shape[0]
        nframes = array.shape[0]
        fc_fr = np.zeros_like(array[0])
        n_fc_rad = rad_dists.shape[0]

        w = int(np.floor(size_fc/2.))
        # fcomp in the center of a zeros frame
        fc_fr[ceny-w:ceny+w+1, cenx-w:cenx+w+1] = psf_template

        array_out = np.zeros_like(array)
        for fr in range(nframes):
            tmp = np.zeros_like(array[0])
            for branch in range(n_branches):
                ang = (branch * 2 * np.pi / n_branches) + np.deg2rad(theta)
                for i in range(n_fc_rad):
                    rad = rad_dists[i]
                    y = rad * np.sin(ang - np.deg2rad(angle_list[fr]))
                    x = rad * np.cos(ang - np.deg2rad(angle_list[fr]))
                    tmp += frame_shift(fc_fr, y, x, imlib=imlib,
                                       interpolation=interpolation) * flevel
                    array_out[fr] = array[fr] + tmp

        if verbose:
            for branch in range(n_branches):
                print('Branch '+str(branch+1)+':')
                for i in range(n_fc_rad):
                    ang = (branch * 2 * np.pi / n_branches) + np.deg2rad(theta)
                    posy = rad_dists[i] * np.sin(ang) + ceny
                    posx = rad_dists[i] * np.cos(ang) + cenx
                    rad_arcs = rad_dists[i]*plsc
                    msg ='\t(X,Y)=({:.2f}, {:.2f}) at {:.2f} arcsec ({:.2f} pxs)'
                    print(msg.format(posx, posy, rad_arcs, rad_dists[i]))

        return array_out

    # ADI+IFS case
    if array.ndim == 4 and psf_template.ndim == 3:
        if psf_template.shape[2] % 2 == 0:
            raise ValueError("Only odd-sized PSF is accepted")
        ceny, cenx = frame_center(array[0,0])
        ceny = int(float(ceny))
        cenx = int(float(cenx))
        rad_dists = np.array(rad_dists)
        if not rad_dists[-1]<array[0].shape[1]/2.:
            msg = 'rad_dists last location is at the border (or outside) of the field'
            raise ValueError(msg)

        sizey = array.shape[2]
        sizex = array.shape[3]
        size_fc = psf_template.shape[2] # considering square frames
        nframes_wav = array.shape[0]
        nframes_adi = array.shape[1]
        fc_fr = np.zeros((nframes_wav, sizey, sizex), dtype=np.float64) # now a 3d cube
        n_fc_rad = rad_dists.shape[0]

        for i in range(nframes_wav):
            w = int(np.floor(size_fc/2.))
            # fcomp in the center of a zeros frame
            if (psf_template[0].shape[1]%2) == 0:
                fc_fr[i, ceny-w:ceny+w, cenx-w:cenx+w] = psf_template[i]
            else:
                fc_fr[i, ceny-w:ceny+w+1, cenx-w:cenx+w+1] = psf_template[i]

        array_out = np.zeros_like(array)

        for fr in range(nframes_adi):
            tmp = np.zeros_like(fc_fr)
            for branch in range(n_branches):
                ang = (branch * 2 * np.pi / n_branches) + np.deg2rad(theta)
                for i in range(n_fc_rad):
                    rad = rad_dists[i]
                    y = rad * np.sin(ang - np.deg2rad(angle_list[fr]))
                    x = rad * np.cos(ang - np.deg2rad(angle_list[fr]))
                    if isinstance(flevel, int) or isinstance(flevel, float):
                        tmp += _cube_shift(fc_fr, y, x, imlib=imlib,
                                           interpolation=interpolation) * flevel
                    else:
                        shift = _cube_shift(fc_fr, y, x, imlib=imlib,
                                            interpolation=interpolation)
                        tmp += [shift[i]*flevel[i] for i in range(len(flevel))]
            array_out[:, fr] = array[:, fr] + tmp

        if verbose:
            for branch in range(n_branches):
                print('Branch '+str(branch+1)+':')
                for i in range(n_fc_rad):
                    ang = (branch * 2 * np.pi / n_branches) + np.deg2rad(theta)
                    rad_arcs = rad_dists[i]*plsc
                    msg ='\t(X,Y)=({:.2f}, {:.2f}) at {:.2f} arcsec ({:.2f} pxs)'
                    posy = rad_dists[i] * np.sin(ang) + ceny
                    posx = rad_dists[i] * np.cos(ang) + cenx
                    print(msg.format(posx, posy, rad_arcs, rad_dists[i]))

    return array_out


def _cube_shift(cube, y, x, imlib, interpolation):
    """ Shifts the X-Y coordinates of a cube or 3D array by x and y values. """
    cube_out = np.zeros_like(cube)
    for i in range(cube.shape[0]):
        cube_out[i] = frame_shift(cube[i], y, x, imlib, interpolation)
    return cube_out


def frame_inject_companion(array, array_fc, pos_y, pos_x, flux,
                           imlib='opencv', interpolation='lanczos4'):
    """ Injects a fake companion in a single frame (it could be a single
     multi-wavelength frame) at given coordinates.
    """
    if not (array.ndim == 2 or array.ndim == 3):
        raise TypeError('Array is not a 2d or 3d array.')
    if array.ndim == 2:
        size_fc = array_fc.shape[0]
        ceny, cenx = frame_center(array)
        ceny = int(ceny)
        cenx = int(cenx)
        fc_fr = np.zeros_like(array)
        w = int(np.floor(size_fc/2.))
        # fcomp in the center of a zeros frame
        fc_fr[ceny-w:ceny+w+1, cenx-w:cenx+w+1] = array_fc
        array_out = array + frame_shift(fc_fr, pos_y-ceny, pos_x-cenx, imlib,
                                        interpolation) * flux

    if array.ndim == 3:
        size_fc = array_fc.shape[1]
        ceny, cenx = frame_center(array[0])
        ceny = int(ceny)
        cenx = int(cenx)
        fc_fr = np.zeros_like(array)
        w = int(np.floor(size_fc/2.))
        # fcomp in the center of a zeros frame
        fc_fr[:, ceny-w:ceny+w+1, cenx-w:cenx+w+1] = array_fc
        array_out = array + _cube_shift(fc_fr, pos_y - ceny, pos_x - cenx,
                                        imlib, interpolation) * flux

    return array_out


def create_psf_template(array, size, fwhm=4, verbose=True, collapse='mean'):
    """ Creates a psf template from a cube of non-saturated off-axis frames of
    the star by taking the mean and normalizing the psf flux.

    Parameters
    ----------
    array : array_like, 3d
        Input cube.
    size : int
        Size of the squared subimage.
    fwhm: float, optional
        The size of the Full Width Half Maximum in pixel.
    verbose : {True,False}, bool optional
        Whether to print to stdout information about file opening, cropping and
        completion of the psf template.
    collapse : {'mean','median'}, string optional
        Defines the way the frames are collapsed.

    Returns
    -------
    psf_normd : array_like
        Normalized PSF.
    """
    if not (array.ndim == 3 or array.ndim == 4):
        raise TypeError('Array is not a cube, 3d or 4d array.')

    n = array.shape[0]
    psf = cube_crop_frames(array, size=size, verbose=verbose)
    if collapse == 'mean':
        psf = np.mean(psf, axis=0)
    elif collapse == 'median':
        psf = np.median(psf, axis=0)
    else:
        raise TypeError('Collapse mode not recognized.')

    psf_normd = psf_norm(psf, size=size, fwhm=fwhm)

    if verbose:
        print("Done scaled PSF template from the average of", n,"frames.")
    return psf_normd


def psf_norm(array, fwhm=4, size=None, threshold=None, mask_core=None,
             imlib='opencv', interpolation='lanczos4', force_odd=True,
             full_output=False, verbose=False):
    """ Scales a PSF (2d or 3d array), so the 1*FWHM aperture flux equals 1.

    Parameters
    ----------
    array: array_like
        The PSF, 2d (ADI data) or 3d array (IFS data).
    fwhm: int or float or 1d array, optional
        The the Full Width Half Maximum in pixels. It can handle a different
        FWHM value for different wavelengths (IFS data).
    size : int or None, optional
        If int it will correspond to the size of the squared subimage to be
        cropped form the psf array.
    threshold : None of float, optional
        Sets to zero small values, trying to leave only the core of the PSF.
    mask_core : None of float, optional
        Sets the radius of a circular aperture for the core of the PSF,
        everything else will be set to zero.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    force_odd : str, optional
        If True the resulting array will have odd size (and the PSF will be
        placed at its center). If False, and the frame size is even, then the
        PSF will be put at the center of an even-sized frame.
    full_output : bool, optional
        If True the flux in a FWHM aperture is returned along with the
        normalized PSF.
    verbose : bool, optional
        If True intermediate results are printed out.

    Returns
    -------
    psf_norm: array_like
        The normalized psf.

    If ``full_output`` is True the flux in a FWHM aperture is returned along
    with the normalized PSF.
    """
    def psf_norm_2d(array, fwhm, size, threshold, mask_core, full_output,
                    verbose):
        """ 2d case """
        if size is not None:
            if size < array.shape[0]:
                psfs = frame_crop(array, size, force=force_odd, verbose=False)
            else:
                psfs = array.copy()
        else:
            psfs = array.copy()

        # we check if the psf is centered and fix it if needed
        cy, cx = frame_center(psfs, verbose=False)
        xcom, ycom = photutils.centroid_com(psfs)
        if not (np.allclose(cy, ycom, atol=1e-2) or
                np.allclose(cx, xcom, atol=1e-2)):
            # first we find the centroid and put it in the center of the array
            centroidy, centroidx = fit_2dgaussian(psfs, fwhmx=fwhm, fwhmy=fwhm)
            shiftx, shifty = centroidx - cx, centroidy - cy
            psfs = frame_shift(array, -shifty, -shiftx, imlib=imlib,
                               interpolation=interpolation)
            if size is not None:
                psfs = frame_crop(psfs, size, force=force_odd, verbose=False)

            for _ in range(2):
                centroidy, centroidx = fit_2dgaussian(psfs, fwhmx=fwhm,
                                                      fwhmy=fwhm)
                cy, cx = frame_center(psfs, verbose=False)
                shiftx, shifty = centroidx - cx, centroidy - cy
                psfs = frame_shift(psfs, -shifty, -shiftx)

        # we check whether the flux is normalized and fix it if needed
        fwhm_aper = photutils.CircularAperture((frame_center(psfs)), fwhm/2.)
        fwhm_aper_phot = photutils.aperture_photometry(psfs, fwhm_aper,
                                                       method='exact')
        fwhm_flux = np.array(fwhm_aper_phot['aperture_sum'])
        if verbose:
            print("Flux in 1xFWHM aperture: {}".format(fwhm_flux))

        if fwhm_flux > 1.1 or fwhm_flux < 0.9:
            psf_norm_array = psfs / np.array(fwhm_aper_phot['aperture_sum'])
        else:
            psf_norm_array = psfs

        if threshold is not None:
            psf_norm_array[np.where(psf_norm_array < threshold)] = 0

        if mask_core is not None:
            psf_norm_array = get_circle(psf_norm_array, radius=mask_core)

        if full_output:
            return psf_norm_array, fwhm_flux
        else:
            return psf_norm_array
    ############################################################################

    if array.ndim == 2:
        y, x = array.shape
        if size is not None:
            if force_odd and size % 2 == 0:
                size += 1
                msg = "`Force_odd` is True therefore `size` was set to {}"
                print(msg.format(size))
        else:
            if force_odd and y % 2 == 0:
                size = y - 1
                msg = "`Force_odd` is True and frame size is even, therefore "
                msg += "new frame size was set to {}"
                print(msg.format(size))
        res = psf_norm_2d(array, fwhm, size, threshold, mask_core, full_output,
                          verbose)

    elif array.ndim == 3:
        y, x = array[0].shape
        if size is not None:
            if force_odd and size % 2 == 0:
                size += 1
                msg = "`Force_odd` is True therefore `size` was set to {}"
                print(msg.format(size))
        else:
            if force_odd and y % 2 == 0:
                size = y - 1
                msg = "`Force_odd` is True and frame size is even, therefore "
                msg += "new frame size was set to {}"
                print(msg.format(size))

        if isinstance(fwhm, (int, float)):
            fwhm = [fwhm for _ in range(array.shape[0])]
        array_out = np.zeros((array.shape[0], size, size))
        if full_output:
            fwhm_flux = np.zeros((array.shape[0]))

        for fr in range(array.shape[0]):
            restemp = psf_norm_2d(array[fr], fwhm[fr], size, threshold,
                                  mask_core, full_output, False)
            if full_output:
                array_out[fr] = restemp[0]
                fwhm_flux[fr] = restemp[1]
            else:
                array_out[fr] = restemp

        if full_output:
            res = (array_out, fwhm_flux)
        else:
            res = array_out

    return res






#! /usr/bin/env python

"""
Module with frame parallactica angles calculations and de-rotation routines for
ADI.
"""
from __future__ import print_function

__author__ = 'V. Christiaens @ UChile/ULg, Carlos Alberto Gomez Gonzalez'
__all__ = ['compute_paral_angles',
           'compute_derot_angles_pa',
           'compute_derot_angles_cd',
           'check_pa_vector']

import math
import numpy as np
import os
from ..fits import open_fits
from astropy.coordinates import FK5
from astropy.coordinates.sky_coordinate import SkyCoord
from astropy.time import Time
from astropy.units import hourangle, degree 


def compute_paral_angles(header, latitude, ra_key, dec_key, lst_key, 
                         acqtime_key, date_key='DATE-OBS'):
    """Calculates the parallactic angle for a frame, taking coordinates and
    local sidereal time from fits-headers (frames taken in an alt-az telescope 
    with the image rotator off).
    
    The coordinates in the header are assumed to be J2000 FK5 coordinates.
    The spherical trigonometry formula for calculating the parallactic angle
    is taken from Astronomical Algorithms (Meeus, 1998).
    
    Parameters
    ----------
    header : dictionary
        Header of current frame.
    latitude : float
        Latitude of the observatory in degrees. The dictionaries in 
        vip_hci/conf/param.py can be used like: latitude=LBT['latitude'].
    ra_key, dec_key, lst_key, acqtime_key, date_key : strings
        Keywords where the values are stored in the header.
        
    Returns
    -------
    pa.value : float
        Parallactic angle in degrees for current header (frame).
    """                                    
    obs_epoch = Time(header[date_key], format='iso', scale='utc')
       
    # equatorial coordinates in J2000
    ra = header[ra_key]                                                         
    dec = header[dec_key]   
    coor = SkyCoord(ra=ra, dec=dec, unit=(hourangle,degree), frame=FK5, 
                    equinox='J2000.0')
    # recalculate for DATE-OBS (precession)
    coor_curr = coor.transform_to(FK5(equinox=obs_epoch))
    
    # new ra and dec in radians
    ra_curr = coor_curr.ra                                                      
    dec_curr = coor_curr.dec
        
    lst_split = header[lst_key].split(':')
    lst = float(lst_split[0])+float(lst_split[1])/60+float(lst_split[2])/3600
    exp_delay = (header[acqtime_key] * 0.5) / 3600
    # solar to sidereal time
    exp_delay = exp_delay*1.0027                                                
    
    # hour angle in degrees
    hour_angle = (lst + exp_delay) * 15 - ra_curr.deg                           
    hour_angle = np.deg2rad(hour_angle)                                         
    latitude = np.deg2rad(latitude)                                             
    
    # PA formula from Astronomical Algorithms 
    pa = -np.rad2deg(np.arctan2(-np.sin(hour_angle), np.cos(dec_curr) * \
                 np.tan(latitude) - np.sin(dec_curr) * np.cos(hour_angle)))     
  
    #if dec_curr.value > latitude:  pa = (pa.value + 360) % 360
    
    return pa.value

def compute_derot_angles_pa(objname_tmp_A,digit_format=3,objname_tmp_B='',
                            inpath='./',writing=False, outpath='./', 
                            list_obj=None, 
                            PosAng_st_key='HIERARCH ESO ADA POSANG',
                            PosAng_nd_key='HIERARCH ESO ADA POSANG END', 
                            verbose=False):
    """
    Function that returns a numpy vector of angles to derotate datacubes so as 
    to match North up, East left, based on the mean of the Position Angle at 
    the beginning and the end of the exposure.
    => It is twice more precise than function derot_angles_CD (there can be 
    >1deg difference in the resulting angle vector returned for fast rotators 
    with long exposures!), but it requires:
    1) a header keyword for both the position angle at start and end of exposure
    2) no skewness of the frames

    The output is in appropriate format for the pca algorithm in the sense that:
    1) all angles of the output are in degrees
    2) all angles of the ouput  are positive
    3) there is no jump of more than 180 deg between consecutive values (e.g. no
    jump like [350deg,355deg,0deg,5deg] 
    => replaced by [350deg,355deg,360deg,365deg])

    Parameters
    ----------
    objname_tmp_A: string
        Contains the common name of the cubes BEFORE the digits
    digit_format: int, optional
        Number of digits in the name of the cube. The digits are supposed to be 
        the only changing part in the name of one cube to another.
    objname_tmp_B: string, optional
        Contains the name of the cubes AFTER the digits
    inpath: string, optional
        Contains the full path of the directory with the data
    writing: bool, optional {False,True}, optional
        True if you want to write the derotation angles in a txt file.
    outpath: string, optional
        Contains the full path of the directory where you want the txt file to 
        be saved.
    list_obj: integer list or 1-D array, optional
        List of the digits corresponding to the cubes to be considered.
        If not provided, the function will consider automatically all the cubes 
        with objname_tmp_A+digit+objname_tmp_B+'.fits' name structure in the 
        provided "inpath".
    PosAng_st_key, PosAng_nd_key: strings, optional
        Name of the keywords to be looked up in the header, to provide the PA 
        from North at start and end of integration.
    verbose: bool, {False,True}, optional
        True if you want more info to be printed.

    Example:
    -------
    If your cubes are: '/home/foo/out_cube_obj_HK_025_000_sorted.fits',
                       '/home/foo/out_cube_obj_HK_025_001_sorted.fits',
                       '/home/foo/out_cube_obj_HK_025_002_sorted.fits', etc,
    the first arguments should be:
                       objname_tmp_A = 'out_cube_obj_HK_025_'
                       digit_format = 3
                       objname_tmp_B = '_sorted'
                       inpath = /home/foo/'

    Returns
    ------_
    angle_list: 1-D array_like
        vector of angles corresponding to the angular difference between the 
        positive y axis and the North in the image.
        sign convention: positive angles in anti-clockwise direction.
        Opposite values are applied when rotating the image to match North up.
    """
   

    posang_st = []
    posang_nd = []

    if list_obj is None:
        list_obj = []
        for ii in range(10**digit_format):
            digits_ii = numberToString(ii,digit_format)
            if os.path.exists(inpath+objname_tmp_A+digits_ii+objname_tmp_B+ \
'.fits'):   
                list_obj.append(ii)
                _, header = open_fits(inpath+objname_tmp_A+ \
                                      digits_ii+objname_tmp_B+ \
                                      '.fits', verbose=False, 
                                      header=True)
                posang_st.append(header[PosAng_st_key])
                posang_nd.append(header[PosAng_nd_key])
    else:
        for ii in list_obj:
            digits_ii = numberToString(ii,digit_format)
            _, header = open_fits(inpath+objname_tmp_A+digits_ii+ \
                                  objname_tmp_B+'.fits', 
                                  verbose=False, header=True)
            posang_st.append(header[PosAng_st_key])
            posang_nd.append(header[PosAng_nd_key])


    # Write the vector containing parallactic angles
    rot = np.zeros(len(list_obj))
    for ii in range(len(list_obj)):
        rot[ii]=0.-(posang_st[ii]+posang_nd[ii])/2.

    # Check and correct to output at the right format
    rot = check_pa_vector(rot,'deg')

    if verbose:
        print("This is the list of angles to be applied: ")
        for ii in range(len(list_obj)):
            print(ii, ' -> ', rot[ii])

    if writing:
        if outpath == '' or outpath is None: outpath=inpath
        f=open(outpath+'Parallactic_angles.txt','w')
        for ii in range(len(list_obj)):
            print(rot[ii], file=f)
        f.close()

    return rot

def compute_derot_angles_cd(objname_tmp_A, digit_format=3,objname_tmp_B='',
                            inpath='./', skew=False, writing=False, 
                            outpath='./', list_obj=None, cd11_key='CD1_1', 
                            cd12_key='CD1_2', cd21_key='CD2_1', 
                            cd22_key='CD2_2', verbose=False):
    """
    Function that returns a numpy vector of angles to derotate datacubes so as 
    to match North up, East left, based on the CD matrix information contained
    in the header.
    In case the PosAng keyword is present in the header and there is no skewness
    between x and y axes, favor the use of function compute_derot_angles_PA 
    (more precise as it averages for the middle of the exposure).
    The output is in appropriate format for the pca algorithm in the sense that:
    1) all angles of the output are in degrees
    2) all angles of the ouput  are positive
    3) there is no jump of more than 180 deg between consecutive values (e.g. no
    jump like [350deg,355deg,0deg,5deg] 
    => replaced by [350deg,355deg,360deg,365deg])
    Parameters:
    -----------
    objname_tmp_A: string
        Contains the common name of the cubes BEFORE the digits
    digit_format: int, optional
        Number of digits in the name of the cube. The digits are supposed to be 
        the only changing part in the name of one cube to another.
    objname_tmp_B: string, optional
        Contains the name of the cubes AFTER the digits
    inpath: string, optional
        Contains the full path of the directory with the data
    skew: bool, {False,True}, optional
        True if you know there is a different rotation between y- and x- axes. 
        The code also detects automatically if there is >1deg skew between y and
        x axes. In case of skewing, 2 vectors of derotation angles are returned:
        one for x and one for y, instead of only one vector.
    writing: bool, {False,True}, optional
        True if you want to write the derotation angles in a txt file.
    outpath: string, opt
        Contains the full path of the directory where you want the txt file to 
        be saved.
    list_obj: integer list or 1-D array, optional
        List of the digits corresponding to the cubes to be considered.
        If not provided, the function will consider automatically all the cubes 
        with objname_tmp_A+digit+objname_tmp_B+'.fits' name structure in the 
        provided "inpath".
    cd11_key,cd12_key,cd21_key,cd22_key: strings, optional
        Name of the keywords to be looked up in the header, to provide the:
        - partial of first axis coordinate w.r.t. x   (cd11_key)
        - partial of first axis coordinate w.r.t. y   (cd12_key)
        - partial of second axis coordinate w.r.t. x  (cd21_key)
        - partial of second axis coordinate w.r.t. y  (cd22_key)
        Default values are the ones in the headers of ESO or HST fits files.
        For more information, go to:
        http://www.stsci.edu/hst/HST_overview/documents/multidrizzle/ch44.html
    verbose: boolean, {False,True}, optional
        True if you want more info to be printed.
    Example:
    -------
    If your cubes are: '/home/foo/out_cube_obj_HK_025_000_sorted.fits',
                       '/home/foo/out_cube_obj_HK_025_001_sorted.fits',
                       '/home/foo/out_cube_obj_HK_025_002_sorted.fits', etc,
    the first arguments should be:
                       objname_tmp_A = 'out_cube_obj_HK_025_'
                       digit_format = 3
                       objname_tmp_B = '_sorted'
                       inpath = '/home/foo/'
    Return:
    -------
    angle_list: 1-D array_like
        vector of angles corresponding to the angular difference between the 
        positive y axis and the North in the image.
        sign convention: positive angles in anti-clockwise direction.
        Opposite values are applied when rotating the image to match North up.
    Note: if skew is set to True, there are 2 angle_list vectors returned; the
        first to rotate the x-axis and the second for the y-axis.
    """

    cd1_1 = []
    cd1_2 = []
    cd2_1 = []
    cd2_2 = []

    if list_obj is None:
        list_obj = []
        for ii in range(10**digit_format):
            digits_ii = numberToString(ii,digit_format)
            obj_str = inpath+objname_tmp_A+digits_ii+objname_tmp_B+'.fits'
            if os.path.exists(obj_str):   
                list_obj.append(ii)
                _, header = open_fits(obj_str, verbose=False, header=True)
                cd1_1.append(header[cd11_key])
                cd1_2.append(header[cd12_key])
                cd2_1.append(header[cd21_key])
                cd2_2.append(header[cd22_key])
    else:
        for ii in list_obj:
            digits_ii = numberToString(ii,digit_format)
            obj_str = inpath+objname_tmp_A+digits_ii+objname_tmp_B+'.fits'
            _, header = open_fits(obj_str, verbose=False, header=True)
            cd1_1.append(header[cd11_key])
            cd1_2.append(header[cd12_key])
            cd2_1.append(header[cd21_key])
            cd2_2.append(header[cd22_key])

    # Determine if it's a right- or left-handed coord system from first cube
    det=cd1_1[0]*cd2_2[0]-cd1_2[0]*cd2_1[0]
    if det<0:  sgn = -1
    else: sgn = 1
    
    # Write the vector containing parallactic angles
    rot = np.zeros(len(list_obj))
    rot2 = np.zeros(len(list_obj))
    for ii in range(len(cd1_1)):
        if cd2_1[ii]==0 and cd1_2[ii]==0:
            rot[ii]=0
            rot2[ii]=0
        else:
            rot[ii]=0.-np.arctan2(sgn*cd1_2[ii],sgn*cd1_1[ii])
            rot2[ii]=0.-np.arctan2(-cd2_1[ii],cd2_2[ii])
            if rot2[ii] < 0:
                rot2[ii] = (2*math.pi)+rot2[ii]
        if np.floor(rot[ii]) != np.floor(rot2[ii]):
            msg = "There is more than 1deg skewness between y and x! "
            msg2 = "Please re-run the function with argument skew=True"
            raise ValueError(msg+msg2)

    # Check and correct to output at the right format
    rot = check_pa_vector(rot,'rad')
    if skew:
        rot2 = check_pa_vector(rot2,'rad')

    if verbose:
        print("This is the list of angles to be applied: ")
        for ii in range(len(cd1_1)):
            print(ii, ' -> ', rot[ii])
            if skew: print('rot2: ', ii, ' -> ', rot2[ii])

    if writing:
        if outpath == '' or outpath is None: outpath=inpath
        f=open(outpath+'Parallactic_angles.txt','w')
        if skew:
            for ii in range(len(cd1_1)):
                print(rot[ii], rot2[ii], file=f)
        else:
            for ii in range(len(cd1_1)):
                print(rot[ii], file=f)
        f.close()

    if skew:
        return rot, rot2
    else:
        return rot


def check_pa_vector(angle_list, unit='deg'):
    """ Checks if the angle list has the right format to avoid any bug in the 
    pca-adi algorithm. The right format complies to 3 criteria:
       1) angles are expressed in degree
       2) the angles are positive
       3) there is no jump of more than 180 deg between consecutive values (e.g.
       no jump like [350deg,355deg,0deg,5deg] => replaced by 
       [350deg,355deg,360deg,365deg])
       
    Parameters
    ----------
    angle_list: 1D-array_like
        Vector containing the derotation angles
    unit: string, {'deg','rad'}, optional
        The unit type of the input angle list
    
    Returns
    -------
    angle_list: 1-D array_like 
        Vector containing the derotation angles (after correction to comply with
        the 3 criteria, if needed)
    """
    angle_list = angle_list.copy()
    if unit != 'rad' and unit != 'deg':
        raise ValueError("The input unit should either be 'deg' or 'rad'")

    npa = angle_list.shape[0]

    for ii in range(npa):
        if unit == 'rad':
            angle_list[ii] = np.rad2deg(angle_list[ii])
        if angle_list[ii] < 0:
            angle_list[ii] = 360+angle_list[ii]

    correct = False
    sorted_rot = np.sort(angle_list)

    # Check if there is a jump > 180deg  within the angle list
    for ii in range(npa-1):
        if abs(sorted_rot[ii+1]-sorted_rot[ii]) > 180:
            correct = True
            break

    # In the previous case, correct for it by adding 360deg to angles < 180deg
    if correct:
        for ii in range(npa):
            if angle_list[ii] < 180:
                angle_list[ii] = 360+angle_list[ii]

    return angle_list


def numberToString(n, digits):
    """ 
    Converts an int in a string according to the number of desired digits
    
    Parameters
    ----------
    n: int
        Number to be converted into string
    digits: int
        Number of characters in the string. If less than the number of 
        digits of n, it is filled with zeros.
    Examples:
    ---------
    >>> numberToString(23, 3)
    '023'
    >>> numberToString(8, 5)
    '00008'
    
    Returns
    -------
    number: string
    The string representing "n", with "digits" characters.
    """
    number = str(n)
    for i in range(digits - len(number)):
        number = "0" + number
    return number



#! /usr/bin/env python

"""
Module with NMF algorithm for ADI.
"""

from __future__ import division 
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['nmf']

import numpy as np
from sklearn.decomposition import NMF
from ..preproc import cube_derotate, cube_collapse
from ..var import prepare_matrix
from ..conf import timing, time_ini


def nmf(cube, angle_list, cube_ref=None, ncomp=1, scaling=None, max_iter=100,
        random_state=None, mask_center_px=None, imlib='opencv',
        interpolation='lanczos4', collapse='median', full_output=False,
        verbose=True, **kwargs):
    """ Non Negative Matrix Factorization for ADI sequences. Alternative to the
    full-frame ADI-PCA processing that does not rely on SVD or ED for obtaining
    a low-rank approximation of the datacube. This function embeds the 
    scikit-learn NMF algorithm solved through coordinate descent method. 
    
    Parameters
    ----------
    cube : array_like, 3d
        Input cube.
    angle_list : array_like, 1d
        Corresponding parallactic angle for each frame.   
    cube_ref : array_like, 3d, optional
        Reference library cube. For Reference Star Differential Imaging.
    ncomp : int, optional
        How many components are used as for low-rank approximation of the 
        datacube.
    scaling : {None, 'temp-mean', 'spat-mean', 'temp-standard', 'spat-standard'}
        With None, no scaling is performed on the input data before SVD. With 
        "temp-mean" then temporal px-wise mean subtraction is done, with 
        "spat-mean" then the spatial mean is subtracted, with "temp-standard" 
        temporal mean centering plus scaling to unit variance is done and with
        "spat-standard" spatial mean centering plus scaling to unit variance is
        performed.  
    max_iter : int optional
        The number of iterations for the coordinate descent solver.
    random_state : int or None, optional
        Controls the seed for the Pseudo Random Number generator.
    mask_center_px : None or int
        If None, no masking is done. If an integer > 1 then this value is the
        radius of the circular mask.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_rotate`` function.
    collapse : {'median', 'mean', 'sum', 'trimmean'}, str optional
        Sets the way of collapsing the frames for producing a final image.
    full_output: boolean, optional
        Whether to return the final median combined image only or with other 
        intermediate arrays.  
    verbose : {True, False}, bool optional
        If True prints intermediate info and timing. 
    kwargs 
        Additional arguments for scikit-learn NMF algorithm.             
             
    Returns
    -------
    If full_output is False the final frame is returned. If True the algorithm
    returns the reshaped NMF components, the reconstructed cube, the residuals,
    the residuals derotated and the final frame.     
    
    """
    array = cube
    if verbose:  start_time = time_ini()
    n, y, x = array.shape
    
    matrix = prepare_matrix(cube, scaling, mask_center_px, mode='fullfr',
                            verbose=verbose)
    matrix += np.abs(matrix.min())
    if cube_ref is not None:
        matrix_ref = prepare_matrix(cube_ref, scaling, mask_center_px,
                                    mode='fullfr', verbose=verbose)
        matrix_ref += np.abs(matrix_ref.min())
           
    mod = NMF(n_components=ncomp, alpha=0, solver='cd', init='nndsvd', 
              max_iter=max_iter, random_state=random_state, **kwargs) 
    
    # H [ncomp, n_pixels]: Non-negative components of the data
    if cube_ref is not None:
        H = mod.fit(matrix_ref).components_
    else:
        H = mod.fit(matrix).components_          
    
    if verbose:  
        print('Done NMF with sklearn.NMF.')
        timing(start_time)
    # W: coefficients [n_frames, ncomp]
    W = mod.transform(matrix)
        
    reconstructed = np.dot(W, H)
    residuals = matrix - reconstructed
               
    array_out = np.zeros_like(array)
    for i in range(n):
        array_out[i] = residuals[i].reshape(y,x)
            
    array_der = cube_derotate(array_out, angle_list, imlib=imlib,
                              interpolation=interpolation)
    frame = cube_collapse(array_der, mode=collapse)
    
    if verbose:  
        print('Done derotating and combining.')
        timing(start_time)
    if full_output:
        return H.reshape(ncomp,y,x), reconstructed.reshape(n,y,x), array_out, array_der, frame 
    else:
        return frame
    
    



"""
Subpackge ``negfc`` contains an ensemble of algorithms for planet position and
flux estimation. It works with as an optimization procedure using a Simplex
approach or coupled with Monte Carlo methods for posterior sampling. Thanks to
this we can get the posterior distributions of the positions and fluxes allowing
us to define proper error-bars for the photometry and position of the companion.
2 possible ways of sampling the posteriors are possible: using ``emcee`` and its
Affine Invariant MCMC or ``nestle`` with a single ellipsoid nested sampling
procedure (much faster).

The main idea of the NegFC is to inject negative fake companions (candidates)
with varying position and flux in order to minimize (in the case of the Simplex)
a function of merit. This function of merit is defined as:

$chi^2 = sum(|I_j|)$,

where $j \in {1,...,N}$ and $N$ the total number of pixels contained in a
circular aperture (4xfwhm) centered on the injection position. This $chi^2$ is
measured on the PCA-processed frame or cube of residuals.

"""
from __future__ import absolute_import

from .simplex_fmerit import *
from .mcmc_sampling import *
from .nested_sampling import *
from .simplex_optim import *
from .utils_negfc import *


#! /usr/bin/env python

"""
Module with ADI helper functions.
"""

from __future__ import division 
from __future__ import print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = []

import numpy as np


def _find_indices(angle_list, frame, thr, nframes=None, out_closest=False,
                  truncate=False, max_frames=200):
    """ Returns the indices to be left in frames library for annular ADI median
    subtraction, LOCI or annular PCA.

    # TODO: find a more pythonic way to to this!

    Parameters
    ----------
    angle_list : array_like, 1d
        Vector of parallactic angle (PA) for each frame.
    frame : int
        Index of the current frame for which we are applying the PA threshold.
    thr : float
        PA threshold.
    nframes : int, optional
        Number of indices to be left. For annular ADI median subtraction,
        where we keep the closest frames (after the PA threshold). If None then
        all the indices are returned (after the PA threshold).
    out_closest : bool, optional
        If True then the function returns the indices of the 2 closest frames.
    truncate : bool, optional
        Useful for annular PCA, when we want to discard too far away frames and
        avoid increasing the computational cost.
    max_frames : int, optional
        Max frames to leave if ``truncate`` is True.

    Returns
    -------
    indices : arrya_like, 1d
        Vector with the indices left.

    If ``out_closest`` is True then the function returns instead:
    index_prev, index_foll
    """
    n = angle_list.shape[0]
    index_prev = 0
    index_foll = frame
    for i in range(0, frame):
        if np.abs(angle_list[frame] - angle_list[i]) < thr:
            index_prev = i
            break
        else:
            index_prev += 1
    for k in range(frame, n):
        if np.abs(angle_list[k] - angle_list[frame]) > thr:
            index_foll = k
            break
        else:
            index_foll += 1

    if out_closest:
        return index_prev, index_foll-1
    else:
        if nframes is not None:
            # For annular ADI median subtraction, returning n_frames closest
            # indices (after PA thresholding)
            window = int(nframes/2)
            ind1 = index_prev-window
            ind1 = max(ind1, 0)
            ind2 = index_prev
            ind3 = index_foll
            ind4 = index_foll+window
            ind4 = min(ind4, n)
            indices = np.array(list(range(ind1, ind2)) +
                               list(range(ind3, ind4)))
        else:
            # For annular PCA, returning all indices (after PA thresholding)
            half1 = range(0, index_prev)
            half2 = range(index_foll, n)

            # This truncation is done on the annuli after 10*FWHM and the goal
            # is to keep min(num_frames/2, 200) in the library after discarding
            # those based on the PA threshold
            if truncate:
                thr = min(int(n / 2), max_frames)
                if frame < thr:
                    half1 = range(max(0, index_prev - int(thr / 2)), index_prev)
                    half2 = range(index_foll,
                                  min(index_foll + thr - len(half1), n))
                else:
                    half2 = range(index_foll, min(n, int(thr / 2 + index_foll)))
                    half1 = range(max(0, index_prev - thr + len(half2)),
                                  index_prev)
            indices = np.array(list(half1) + list(half2))

        return indices


def _compute_pa_thresh(ann_center, fwhm, delta_rot=1):
    """ Computes the parallactic angle theshold[degrees]
    Replacing approximation: delta_rot * (fwhm/ann_center) / np.pi * 180
    """
    return np.rad2deg(2 * np.arctan(delta_rot * fwhm / (2 * ann_center)))


def _define_annuli(angle_list, ann, n_annuli, fwhm, radius_int, annulus_width,
                   delta_rot, n_segments, verbose):
    """ Function that defines the annuli geometry using the input parameters.
    Returns the parallactic angle threshold, the inner radius and the annulus
    center for each annulus.
    """
    if ann == n_annuli - 1:
        inner_radius = radius_int + (ann * annulus_width - 1)
    else:
        inner_radius = radius_int + ann * annulus_width
    ann_center = (inner_radius + (annulus_width / 2.0))
    pa_threshold = _compute_pa_thresh(ann_center, fwhm, delta_rot)

    mid_range = np.abs(np.amax(angle_list) - np.amin(angle_list)) / 2
    if pa_threshold >= mid_range - mid_range * 0.1:
        new_pa_th = float(mid_range - mid_range * 0.1)
        if verbose:
            msg = 'PA threshold {:.2f} is too big, will be set to {:.2f}'
            print(msg.format(pa_threshold, new_pa_th))
        pa_threshold = new_pa_th

    if verbose:
        msg2 = 'Annulus {:}, PA thresh = {:.2f}, Inn radius = {:.2f}, '
        msg2 += 'Ann center = {:.2f}, N segments = {:} '
        print(msg2.format(int(ann + 1), pa_threshold, inner_radius, ann_center,
                          n_segments))
    return pa_threshold, inner_radius, ann_center


#! /usr/bin/env python

"""
Module with sigma clipping functions.
"""

from __future__ import division, print_function

__author__ = 'Carlos Alberto Gomez Gonzalez', 'V. Christiaens'
__all__ = ['clip_array',
           'sigma_filter']

import numpy as np
from scipy.ndimage.filters import generic_filter
from astropy.stats import median_absolute_deviation


# TODO: If possible, replace this function using
# scipy.ndimage.filters.generic_filter and astropy.stats.sigma_clip
def sigma_filter(frame_tmp, bpix_map, neighbor_box=3, min_neighbors=3,
                 verbose=False):
    """Sigma filtering of pixels in a 2d array.
    
    Parameters
    ----------
    frame_tmp : array_like 
        Input 2d array, image.
    bpix_map: array_like
        Input array of the same size as frame_tmp, indicating the locations of 
        bad/nan pixels by 1 (the rest of the array is set to 0)
    neighbor_box : int, optional
        The side of the square window around each pixel where the sigma and 
        median are calculated.
    min_neighbors : int, optional
        Minimum number of good neighboring pixels to be able to correct the 
        bad/nan pixels
    verbose : bool, optional
        Prints out the number of iterations.
        
    Returns
    -------
    frame_corr : array_like
        Output array with corrected bad/nan pixels
    
    """
    if not frame_tmp.ndim == 2:
        raise TypeError('Input array is not a frame or 2d array')

    sz_y = frame_tmp.shape[0]  # get image y-dim
    sz_x = frame_tmp.shape[1]  # get image x-dim
    bp = bpix_map.copy()       # temporary bpix map; important to make a copy!
    im = frame_tmp             # corrected image
    nb = int(np.sum(bpix_map)) # number of bad pixels remaining
    #In each iteration, correct only the bpix with sufficient good 'neighbors'
    nit = 0                                 # number of iterations
    while nb > 0:
        nit += 1
        wb = np.where(bp)                   # find bad pixels
        gp = 1 - bp                         # temporary good pixel map
        for n in range(nb):
            #0/ Determine the box around each pixel
            half_box = np.floor(neighbor_box/2.)
            hbox_b = min(half_box,wb[0][n])        # half size of the box at the
                                                   # bottom of the pixel
            hbox_t = min(half_box,sz_y-1-wb[0][n]) # half size of the box at the
                                                   # top of the pixel
            hbox_l = min(half_box,wb[1][n])        # half size of the box to the
                                                   # left of the pixel
            hbox_r = min(half_box,sz_x-1-wb[1][n]) # half size of the box to the
                                                   # right of the pixel
            # but in case we are at an edge, we want to extend the box by one 
            # row/column of pixels in the direction opposite to the edge to 
            # have 9 px instead of 6: 
            if half_box == 1:
                if wb[0][n] == sz_y-1:
                    hbox_b = hbox_b+1
                elif wb[0][n] == 0:
                    hbox_t = hbox_t+1
                if wb[1][n] == sz_x-1:
                    hbox_l = hbox_l+1
                elif wb[1][n] == 0:
                    hbox_r = hbox_r+1

            sgp = gp[int(wb[0][n]-hbox_b): int(wb[0][n]+hbox_t+1),
                     int(wb[1][n]-hbox_l): int(wb[1][n]+hbox_r+1)]
            if int(np.sum(sgp)) >= min_neighbors:
                sim = im[int(wb[0][n]-hbox_b): int(wb[0][n]+hbox_t+1),
                         int(wb[1][n]-hbox_l): int(wb[1][n]+hbox_r+1)]
                im[wb[0][n],wb[1][n]] = np.median(sim[np.where(sgp)])
                bp[wb[0][n],wb[1][n]] = 0
        nb = int(np.sum(bp))
    if verbose:
        print('Required number of iterations in the sigma filter: ', nit)
    return im


# TODO: If possible, replace this function using astropy.stats.sigma_clip
def clip_array(array, lower_sigma, upper_sigma, out_good=False, neighbor=False,
               num_neighbor=None, mad=False):
    """Sigma clipping for detecting outlying values in 2d array. If the parameter
    'neighbor' is True the clipping can be performed in a local patch around 
    each pixel, whose size depends on 'neighbor' parameter.
    
    Parameters
    ----------
    array : array_like 
        Input 2d array, image.
    lower_sigma : float 
        Value for sigma, lower boundary.
    upper_sigma : float 
        Value for sigma, upper boundary.
    out_good : bool, optional
        For choosing different outputs.
    neighbor : bool optional
        For clipping over the median of the contiguous pixels.
    num_neighbor : int, optional
        The side of the square window around each pixel where the sigma and 
        median are calculated. 
    mad : bool, optional
        If True, the median absolute deviation will be used instead of the 
        standard deviation.
        
    Returns
    -------
    good : array_like
        If out_good argument is true, returns the indices of not-outlying px.
    bad : array_like 
        If out_good argument is false, returns a vector with the outlier px.
    
    """
    if not array.ndim == 2:
        raise TypeError("Input array is not two dimensional (frame)\n")

    values = array.copy()
    if neighbor and num_neighbor:
        median = generic_filter(array, function=np.median, 
                                size=(num_neighbor,num_neighbor), mode="mirror")
        if mad:
            sigma = generic_filter(array, function=median_absolute_deviation,                                 
                                   size=(num_neighbor,num_neighbor), 
                                   mode="mirror")
        else:
            sigma = generic_filter(array, function=np.std,                                 
                                   size=(num_neighbor,num_neighbor), 
                                   mode="mirror")
    else:
        median = np.median(values)
        sigma = values.std()
        
    good1 = values > (median - lower_sigma * sigma) 
    good2 = values < (median + upper_sigma * sigma)
    bad1 = values < (median - lower_sigma * sigma)
    bad2 = values > (median + upper_sigma * sigma)
    
    if out_good:
        good = np.where(good1 & good2)                                          # normal px indices in both good1 and good2
        return good
    else:
        bad = np.where(bad1 | bad2)                                             # deviating px indices in either bad1 or bad2
        return bad
    
  



#! /usr/bin/env python

"""Module with a dictionary and variables for storing constant parameters.

Usage
-----
from param import VLT_NACO
VLT_NACO['diam']

"""

VLT_SPHERE = {
    'latitude' : -24.627,
    'longitude' : -70.404,
    'plsc' : 0.01225,                       # plate scale [arcsec]/px
    'diam': 8.0,                            # telescope diameter [m]
    }

VLT_NACO = {
    'latitude' : -24.627,
    'longitude' : -70.404,
    'plsc': 0.027190,                       # plate scale [arcsec]/px
    'diam': 8.0,                            # telescope diameter [m]
    'lambdal' : 3.8e-6,                     # filter central wavelength [m] L band
    'camera_filter' : 'Ll',
    # header keywords
    'kw_categ' : 'HIERARCH ESO DPR CATG',   # header keyword for calibration frames: 'CALIB' / 'SCIENCE'
    'kw_type' : 'HIERARCH ESO DPR TYPE'     # header keyword for frame type: 'FLAT,SKY' / 'DARK' / 'OBJECT' / 'SKY'
    }

VLT_SINFONI = {
    'latitude' : -24.627,
    'longitude' : -70.404,
    # plsc: depending on the chosen mode, can also be: 0.125'' or 0.05''
    'plsc': 0.0125,                         # plate scale [arcsec]/px
    'diam': 8.2,                            # telescope diameter [m]
    'camera_filter' : 'H+K',
    'lambdahk': 1.95e-6,                    # wavelength of the middle of H+K
    'lambdah': 1.65e-6,                     # wavelength of the middle of H band
    'lambdak': 2.166e-6,                    # wavelength of the Brackett gamma line (~middle of the K band)
    'spec_res':5e-4,                        # spectral resolution in um (for H+K)
    # header keywords
    'kw_categ' : 'HIERARCH ESO DPR CATG',   # header keyword for calibration frames: 'CALIB' / 'SCIENCE'
    'kw_type' : 'HIERARCH ESO DPR TYPE'     # header keyword for frame type: 'FLAT,SKY' / 'DARK' / 'OBJECT' / 'SKY'
    }

LBT = {
    'latitude' : 32.70131,                  # LBT's latitude in degrees
    'longitude' : -109.889064,              # LBT's longitude in degrees
    'lambdal' : 3.47e-6,                    # central wavelenght L cont2' band [m]
    'plsc' : 0.0106,                        # plate scale [arcsec]/px
    'diam' : 8.4,                           # telescope diameter [m]
    # header keywords
    'lst': 'LBT_LST',                       # Local sidereal Time header keyword
    'ra' : 'LBT_RA',                        # right ascension header keyword
    'dec' : 'LBT_DEC',                      # declination header keyword
    'altitude' : 'LBT_ALT',                 # altitude header keyword
    'azimuth' : 'LBT_AZ',                   # azimuth header keyword
    'exptime' : 'EXPTIME',                  # nominal total integration time per pixel header keyword
    'acqtime' : 'ACQTIME',                  # total controller acquisition time header keyword    
    'filter' : 'LMIR_FW2'                   # filter
    }

KECK_NIRC2 = {
    'plsc_narrow' : 0.009942,               # plate scale [arcsec]/px, narrow camera
    'plsc_medium' : 0.019829,               # plate scale [arcsec]/px, medium camera
    'plsc_wide' : 0.039686,                 # plate scale [arcsec]/px, wide camera   
    'latitude' : 19.82636,                  # Keck's latitude in degrees
    # header keywords
    'camera_name' : 'CAMNAME'               # camera name bwt narrow, medium and wide
    }





#! /usr/bin/env python

"""
Various stat functions.
"""

from __future__ import division, print_function

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['descriptive_stats']

import numpy as np
from matplotlib.pyplot import boxplot


def descriptive_stats(array, verbose=True, label='', mean=False, plot=False):
    """ Simple statistics from vector.
    """
    if mean:  mean = np.mean(array)
    median = np.median(array)
    mini = np.min(array)
    maxi = np.max(array)
    first_qu = np.percentile(array, 25)
    third_qu = np.percentile(array, 75)
    
    if verbose:
        msg = label
        if mean:
            msg += 'min={:.1f} / 1st QU={:.1f} / ave={:.1f} / med={:.1f} / 3rd QU={:.1f} / max={:.1f}'
            print(msg.format(mini, first_qu, mean, median, third_qu, maxi))
        else:
            msg += 'min={:.1f} / 1st QU={:.1f} / med={:.1f} / 3rd QU={:.1f} / max={:.1f}'
            print(msg.format(mini, first_qu, median, third_qu, maxi))
    
    if plot:
        boxplot(array, vert=False, meanline=mean, showfliers=True, sym='.') #whis=range)
    
    if mean:
        return mini, first_qu, mean, median, third_qu, maxi
    else:
        return mini, first_qu, median, third_qu, maxi